from typing import Optional, cast

from isar.eventhandlers.eventhandler import (
    EventHandlerMapping,
    State,
    TimeoutHandlerMapping,
)
from isar.models.events import EmptyMessage
from isar.state_machine.state_machine import StateMachine
from isar.state_machine.states.await_next_mission import AwaitNextMission
from isar.state_machine.states.home import Home
from isar.state_machine.states.monitor import Monitor
from isar.state_machine.states.pausing_return_home import PausingReturnHome
from isar.state_machine.states.resuming_return_home import ResumingReturnHome
from isar.state_machine.states.returning_home import ReturningHome
from isar.state_machine.states.stopping_return_home import StoppingReturnHome
from robot_interface.models.exceptions.robot_exceptions import ErrorMessage, ErrorReason
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.task import ReturnToHome


def test_transitioning_to_returning_home_from_stopping_when_return_home_failed(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.shared_state.robot_battery_level.trigger_event(80.0)
    example_mission: Mission = Mission(
        name="Dummy return home misson", tasks=[ReturnToHome()]
    )
    sync_state_machine.current_state = StoppingReturnHome(
        sync_state_machine, example_mission
    )

    stopping_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        stopping_state.get_event_handler_by_name("successful_stop_event")
    )

    assert event_handler is not None

    transition = event_handler.handler(EmptyMessage())
    sync_state_machine.current_state = transition(sync_state_machine)

    assert type(sync_state_machine.current_state) is Monitor


def test_transition_from_pausing_return_home_to_returning_home(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.current_state = PausingReturnHome(sync_state_machine)

    pausing_return_home_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        pausing_return_home_state.get_event_handler_by_name("failed_pause_event")
    )

    assert event_handler is not None

    error_event = ErrorMessage(
        error_reason=ErrorReason.RobotUnknownErrorException, error_description=""
    )
    transition = event_handler.handler(error_event)

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is ReturningHome


def test_transition_from_resuming_return_home_to_returning_home_state(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.current_state = ResumingReturnHome(sync_state_machine)

    resuming_return_home_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        resuming_return_home_state.get_event_handler_by_name("successful_resume_event")
    )

    assert event_handler is not None

    transition = event_handler.handler(EmptyMessage())

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is ReturningHome


def test_transition_from_returning_home_to_home_robot_status_not_updated(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.shared_state.robot_battery_level.trigger_event(90.0)
    sync_state_machine.current_state = ReturningHome(sync_state_machine)

    returning_home_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        returning_home_state.get_event_handler_by_name("mission_succeeded_event")
    )

    assert event_handler is not None

    transition = event_handler.handler(EmptyMessage())

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is Home
    assert (
        not sync_state_machine.events.robot_service_events.robot_status_changed.check()
    )

    home_state: State = cast(State, sync_state_machine.current_state)
    event_handler_robot_status: Optional[EventHandlerMapping] = (
        home_state.get_event_handler_by_name("robot_status_event")
    )

    assert event_handler_robot_status is not None

    assert not event_handler_robot_status.event.has_event()


def test_return_home_not_cancelled_when_battery_is_low(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.shared_state.robot_battery_level.trigger_event(10.0)
    events = sync_state_machine.events
    sync_state_machine.current_state = ReturningHome(sync_state_machine)

    returning_home_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        returning_home_state.get_event_handler_by_name("start_mission_event")
    )

    assert event_handler is not None

    transition = event_handler.handler(EmptyMessage())

    assert transition is None
    assert events.api_requests.start_mission.response.has_event()
    start_mission_event_response = events.api_requests.start_mission.response.check()
    assert not start_mission_event_response.mission_started


def test_return_home_starts_when_battery_is_low(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.shared_state.robot_battery_level.trigger_event(10.0)
    sync_state_machine.current_state = AwaitNextMission(sync_state_machine)

    await_next_mission_state: State = cast(State, sync_state_machine.current_state)
    timer: Optional[TimeoutHandlerMapping] = (
        await_next_mission_state.get_event_timer_by_name("should_return_home_timer")
    )

    assert timer is not None

    transition = timer.handler()

    sync_state_machine.current_state = transition(sync_state_machine)

    assert type(sync_state_machine.current_state) is ReturningHome



import time
from collections import deque

from pytest_mock import MockerFixture

from isar.robot.robot_status import RobotStatusThread
from isar.state_machine.states_enum import States
from tests.test_mocks.robot_interface import StubRobotBlockedProtectiveStopToHomeTest
from tests.test_mocks.state_machine_mocks import (
    RobotServiceThreadMock,
    StateMachineThreadMock,
)


def test_state_machine_idle_to_blocked_protective_stop_to_idle(
    state_machine_thread: StateMachineThreadMock,
    robot_service_thread: RobotServiceThreadMock,
    mocker: MockerFixture,
) -> None:
    # Robot status check happens every 5 seconds by default, so we mock the behavior
    # to poll for status imediately
    mocker.patch.object(
        RobotStatusThread, "_is_ready_to_poll_for_status", return_value=True
    )

    robot_service_thread.robot_service.robot = StubRobotBlockedProtectiveStopToHomeTest(
        robot_service_thread.robot_service.shared_state.state
    )

    state_machine_thread.start()
    robot_service_thread.start()
    time.sleep(5)

    assert state_machine_thread.state_machine.transitions_list == deque(
        [States.UnknownStatus, States.BlockedProtectiveStop, States.Home]
    )



import logging
import os
import time
from queue import Empty, Queue
from typing import Any, Tuple

import backoff
from paho.mqtt import client as mqtt
from paho.mqtt.client import Client
from paho.mqtt.packettypes import PacketTypes
from paho.mqtt.properties import Properties

from isar.config.settings import settings
from robot_interface.telemetry.mqtt_client import MqttClientInterface


def props_expiry(seconds: int) -> Properties:
    p = Properties(PacketTypes.PUBLISH)
    p.MessageExpiryInterval = seconds
    return p


def _on_success(data: dict) -> None:
    logging.getLogger("mqtt_client").info("Connected to MQTT Broker")
    logging.getLogger("mqtt_client").debug(
        f"Elapsed time: {data['elapsed']}, Tries: {data['tries']}"
    )


def _on_backoff(data: dict) -> None:
    logging.getLogger("mqtt_client").warning(
        f"Failed to connect, retrying in {data['wait']} seconds"
    )


def _on_giveup(data: dict) -> None:
    logging.getLogger("mqtt_client").error(
        "Failed to connect to MQTT Broker within set backoff strategy."
    )


class MqttClient(MqttClientInterface):
    def __init__(self, mqtt_queue: Queue) -> None:
        self.logger = logging.getLogger("mqtt_client")
        self.logger.setLevel("INFO")
        self.mqtt_queue: Queue = mqtt_queue

        username: str = settings.MQTT_USERNAME
        password: str = ""
        try:
            password = os.environ["ISAR_MQTT_PASSWORD"]
        except KeyError:
            self.logger.warning(
                "Failed to retrieve ISAR_MQTT_PASSWORD from environment. Attempting "
                "with empty string as password."
            )

        self.host: str = settings.MQTT_HOST

        # Fix for mqtt running on localhost in docker
        if "IS_DOCKER" in os.environ and self.host == "localhost":
            self.host = "host.docker.internal"

        self.port: int = settings.MQTT_PORT

        self.client: Client = Client(
            protocol=mqtt.MQTTv5, callback_api_version=mqtt.CallbackAPIVersion.VERSION2
        )

        self.client.enable_logger(logger=self.logger)

        dirname = os.path.dirname(__file__)

        if settings.MQTT_SSL_ENABLED:
            cert_path = os.path.join(dirname, "../../../config/certs/ca-cert.pem")
            self.client.tls_set(ca_certs=cert_path)

        self.client.on_connect = self.on_connect
        self.client.on_disconnect = self.on_disconnect

        self.client.username_pw_set(username=username, password=password)

    def run(self) -> None:
        self.connect(host=self.host, port=self.port)
        self.client.loop_start()

        while True:
            if not self.client.is_connected():
                time.sleep(0)  # avoid CPU spin
                continue
            try:
                item = self.mqtt_queue.get(timeout=1)
                if len(item) == 4:
                    topic, payload, qos, retain = item
                    properties = None
                else:
                    topic, payload, qos, retain, properties = item
            except Empty:
                continue

            self.publish(
                topic=topic,
                payload=payload,
                qos=qos,
                retain=retain,
                properties=properties,
            )

    def on_connect(
        self, client: Any, userdata: Any, flags: Any, reason_code: str, properties: Any
    ) -> None:
        self.logger.info(f"Connected: {reason_code}")

    def on_disconnect(self, client: Any, userdata: Any, *args: Tuple[Any]) -> None:
        if not args:
            return
        reason_code = args[0] if len(args) < 3 else args[1]
        rc = getattr(reason_code, "value", reason_code)
        if rc != 0:
            self.logger.warning(f"Unexpected disconnect: {reason_code}.")

    @backoff.on_exception(
        backoff.expo,
        ConnectionRefusedError,
        max_time=300,
        on_success=_on_success,
        on_backoff=_on_backoff,
        on_giveup=_on_giveup,
    )
    def connect(self, host: str, port: int) -> None:
        self.logger.info("Attempting to connect to MQTT Broker")
        self.logger.info("Host: %s, Port: %s", host, port)
        self.client.connect(host=host, port=port)

    def publish(
        self,
        topic: str,
        payload: str,
        qos: int = 0,
        retain: bool = False,
        properties: Properties = None,
    ) -> None:
        self.logger.debug("Publishing message to topic: %s", topic)
        self.client.publish(
            topic=topic, payload=payload, qos=qos, retain=retain, properties=properties
        )



import logging
import time
from threading import Event, Thread

from isar.config.settings import settings
from isar.models.events import (
    EmptyMessage,
    RobotServiceEvents,
    SharedState,
    StateMachineEvents,
)
from robot_interface.models.exceptions.robot_exceptions import RobotException
from robot_interface.robot_interface import RobotInterface


class RobotStatusThread(Thread):
    def __init__(
        self,
        robot: RobotInterface,
        signal_thread_quitting: Event,
        shared_state: SharedState,
        robot_service_events: RobotServiceEvents,
        state_machine_events: StateMachineEvents,
    ):
        self.logger = logging.getLogger("robot")
        self.shared_state: SharedState = shared_state
        self.robot_service_events: RobotServiceEvents = robot_service_events
        self.state_machine_events: StateMachineEvents = state_machine_events
        self.robot: RobotInterface = robot
        self.signal_thread_quitting: Event = signal_thread_quitting
        self.robot_status_poll_interval: float = settings.ROBOT_API_STATUS_POLL_INTERVAL
        self.last_robot_status_poll_time: float = time.time()
        Thread.__init__(self, name="Robot status thread")

    def stop(self) -> None:
        return

    def _is_ready_to_poll_for_status(self) -> bool:
        time_since_last_robot_status_poll = (
            time.time() - self.last_robot_status_poll_time
        )
        return time_since_last_robot_status_poll > self.robot_status_poll_interval

    def run(self) -> None:
        if self.signal_thread_quitting.is_set():
            return

        thread_check_interval = settings.THREAD_CHECK_INTERVAL

        while not self.signal_thread_quitting.wait(thread_check_interval):

            if not self._is_ready_to_poll_for_status():
                continue

            try:
                self.last_robot_status_poll_time = time.time()

                robot_status = self.robot.robot_status()

                if robot_status is not self.shared_state.robot_status.check():
                    self.shared_state.robot_status.update(robot_status)
                    self.robot_service_events.robot_status_changed.trigger_event(
                        EmptyMessage()
                    )
            except RobotException as e:
                self.logger.error(f"Failed to retrieve robot status: {e}")
                continue
            except Exception as e:
                self.logger.error(
                    f"Unhandled exception in robot status service: {str(e)}"
                )
                continue
        self.logger.info("Exiting robot status thread")



from typing import TYPE_CHECKING, List, Union

import isar.state_machine.states.await_next_mission as AwaitNextMission
import isar.state_machine.states.monitor as Monitor
import isar.state_machine.states.returning_home as ReturningHome
from isar.apis.models.models import ControlMissionResponse
from isar.eventhandlers.eventhandler import EventHandlerMapping, State, Transition
from isar.models.events import EmptyMessage
from isar.state_machine.states_enum import States
from robot_interface.models.exceptions.robot_exceptions import ErrorMessage

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class Stopping(State):

    def __init__(self, state_machine: "StateMachine", mission_id: str):
        events = state_machine.events

        def _failed_stop_event_handler(
            error_message: ErrorMessage,
        ) -> Transition[Monitor.Monitor]:
            state_machine.logger.warning(
                f"Failed to stop mission: {error_message.error_description}"
            )
            return Monitor.transition_with_existing_mission(mission_id)

        def _successful_stop_event_handler(
            successful_stop: EmptyMessage,
        ) -> Union[
            Transition[AwaitNextMission.AwaitNextMission],
            Transition[ReturningHome.ReturningHome],
        ]:
            if not state_machine.battery_level_is_above_mission_start_threshold():
                return ReturningHome.transition_and_start_mission()
            return AwaitNextMission.transition()

        event_handlers: List[EventHandlerMapping] = [
            EventHandlerMapping[ErrorMessage](
                name="failed_stop_event",
                event=events.robot_service_events.mission_failed_to_stop,
                handler=_failed_stop_event_handler,
            ),
            EventHandlerMapping[EmptyMessage](
                name="successful_stop_event",
                event=events.robot_service_events.mission_successfully_stopped,
                handler=_successful_stop_event_handler,
            ),
        ]
        super().__init__(
            state_name=States.Stopping,
            state_machine=state_machine,
            event_handler_mappings=event_handlers,
        )


def transition_and_trigger_stop(
    mission_id: str, should_respond_to_API_request: bool = False
) -> Transition[Stopping]:
    def _transition(state_machine: "StateMachine") -> Stopping:
        state_machine.events.state_machine_events.stop_mission.trigger_event(
            EmptyMessage()
        )
        if should_respond_to_API_request:
            state_machine.events.api_requests.stop_mission.response.trigger_event(
                ControlMissionResponse(success=True)
            )
        return Stopping(state_machine, mission_id)

    return _transition



import time
from datetime import datetime
from typing import Tuple

from alitra import Frame, Orientation, Pose, Position

from isar.modules import ApplicationContainer
from isar.storage.uploader import Uploader
from robot_interface.models.inspection.inspection import (
    ImageMetadata,
    Inspection,
    InspectionBlob,
)
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.task import TakeImage
from tests.test_mocks.blob_storage import StorageEmptyBlobPathsFake, StorageFake
from tests.test_mocks.mqtt_client import MqttPublisherFake
from tests.test_mocks.state_machine_mocks import UploaderThreadMock

MISSION_ID = "some-mission-id"
ARBITRARY_IMAGE_METADATA = ImageMetadata(
    start_time=datetime.now(),
    robot_pose=Pose(
        Position(0, 0, 0, Frame("asset")),
        Orientation(x=0, y=0, z=0, w=1, frame=Frame("asset")),
        Frame("asset"),
    ),
    target_position=Position(0, 0, 0, Frame("asset")),
    file_type="jpg",
)


def test_should_upload_from_queue(
    container: ApplicationContainer, uploader_thread: UploaderThreadMock
) -> None:
    uploader_thread.start()

    take_image_task = TakeImage()
    mission: Mission = Mission(name="Dummy misson", tasks=[take_image_task])

    assert isinstance(mission.tasks[0], TakeImage)
    inspection = InspectionBlob(
        metadata=ARBITRARY_IMAGE_METADATA, id=mission.tasks[0].inspection_id
    )

    message: Tuple[Inspection, Mission] = (
        inspection,
        mission,
    )

    uploader: Uploader = container.uploader()

    uploader.upload_queue.put(message)
    time.sleep(0.01)

    storage_handler: StorageFake = uploader.storage_handlers[0]  # type: ignore
    assert inspection in storage_handler.stored_inspections


def test_should_retry_failed_upload_from_queue(
    container: ApplicationContainer, uploader_thread: UploaderThreadMock
) -> None:
    uploader_thread.start()

    INSPECTION_ID = "123-456"
    inspection = InspectionBlob(metadata=ARBITRARY_IMAGE_METADATA, id=INSPECTION_ID)
    mission: Mission = Mission(name="Dummy Mission")

    message: Tuple[Inspection, Mission] = (
        inspection,
        mission,
    )

    uploader: Uploader = container.uploader()
    storage_handler: StorageFake = uploader.storage_handlers[0]  # type: ignore

    # Need it to fail so that it retries
    storage_handler.will_fail = True
    uploader.upload_queue.put(message)
    time.sleep(1)

    # Should not upload, instead raise StorageException
    assert not storage_handler.blob_exists(inspection)
    storage_handler.will_fail = False
    time.sleep(3)

    # After some time, it should have retried and now it should be successful
    assert storage_handler.blob_exists(inspection)


def test_should_not_publish_when_blob_paths_are_empty(
    container: ApplicationContainer, uploader_thread: UploaderThreadMock
) -> None:
    uploader_thread.start()

    mission: Mission = Mission(name="Dummy mission")
    inspection: Inspection = InspectionBlob(
        metadata=ARBITRARY_IMAGE_METADATA, id="blob-empty"
    )

    uploader: Uploader = container.uploader()

    storage_handler: StorageEmptyBlobPathsFake() = StorageEmptyBlobPathsFake()  # type: ignore
    uploader.storage_handlers[0] = storage_handler

    mqtt_fake = MqttPublisherFake()
    uploader.mqtt_publisher = mqtt_fake

    message: Tuple[Inspection, Mission] = (
        inspection,
        mission,
    )
    uploader.upload_queue.put(message)
    time.sleep(1)

    assert inspection in storage_handler.stored

    assert len(mqtt_fake.published) == 0



from typing import Optional, cast

from isar.eventhandlers.eventhandler import EventHandlerMapping, State
from isar.models.events import EmptyMessage
from isar.state_machine.state_machine import StateMachine
from isar.state_machine.states.paused import Paused
from isar.state_machine.states.pausing import Pausing
from isar.state_machine.states.stopping_paused_mission import StoppingPausedMission


def test_transition_from_pausing_to_paused(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.current_state = Pausing(sync_state_machine, "mission_id")

    pausing_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        pausing_state.get_event_handler_by_name("successful_pause_event")
    )

    assert event_handler is not None

    transition = event_handler.handler(EmptyMessage())

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is Paused


def test_transition_from_paused_to_stopping_paused_mission(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.current_state = Paused(sync_state_machine, "test_id")

    paused_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        paused_state.get_event_handler_by_name("stop_mission_event")
    )

    assert event_handler is not None

    transition = event_handler.handler("test_id")

    sync_state_machine.current_state = transition(sync_state_machine)

    assert type(sync_state_machine.current_state) is StoppingPausedMission
    assert sync_state_machine.events.api_requests.stop_mission.response.has_event()


def test_stop_request_with_wrong_id_in_paused(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.current_state = Paused(sync_state_machine, "test_id")

    paused_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        paused_state.get_event_handler_by_name("stop_mission_event")
    )

    assert event_handler is not None

    transition = event_handler.handler("wrong_test_id")

    assert transition is None
    assert sync_state_machine.events.api_requests.stop_mission.response.has_event()



from enum import Enum

from pydantic import BaseModel


class MediaConnectionType(str, Enum):
    LiveKit = "LiveKit"


class MediaConfig(BaseModel):
    url: str
    token: str
    media_connection_type: MediaConnectionType



from typing import TYPE_CHECKING, List, Optional

import isar.state_machine.states.await_next_mission as AwaitNextMission
import isar.state_machine.states.pausing as Pausing
import isar.state_machine.states.stopping as Stopping
import isar.state_machine.states.stopping_due_to_maintenance as StoppingDueToMaintenance
import isar.state_machine.states.stopping_go_to_lockdown as StoppingGoToLockdown
import isar.state_machine.states.stopping_go_to_recharge as StoppingGoToRecharge
from isar.apis.models.models import ControlMissionResponse, MissionStartResponse
from isar.config.settings import settings
from isar.eventhandlers.eventhandler import EventHandlerMapping, State, Transition
from isar.models.events import EmptyMessage
from isar.state_machine.states_enum import States
from robot_interface.models.exceptions.robot_exceptions import ErrorMessage
from robot_interface.models.mission.mission import Mission

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class Monitor(State):

    def __init__(self, state_machine: "StateMachine", mission_id: str):
        events = state_machine.events
        shared_state = state_machine.shared_state

        def _pause_mission_event_handler(
            should_pause: EmptyMessage,
        ) -> Transition[Pausing.Pausing]:
            state_machine.events.api_requests.pause_mission.response.trigger_event(
                ControlMissionResponse(success=True)
            )
            state_machine.events.state_machine_events.pause_mission.trigger_event(
                EmptyMessage()
            )
            return Pausing.transition(mission_id)

        def _robot_battery_level_updated_handler(
            battery_level: float,
        ) -> Optional[Transition[StoppingGoToRecharge.StoppingGoToRecharge]]:
            if (
                battery_level is None
                or battery_level >= settings.ROBOT_MISSION_BATTERY_START_THRESHOLD
            ):
                return None

            state_machine.logger.warning(
                "Cancelling current mission due to low battery"
            )
            state_machine.events.state_machine_events.stop_mission.trigger_event(
                EmptyMessage()
            )
            return StoppingGoToRecharge.transition(mission_id)

        def _send_to_lockdown_event_handler(
            should_lockdown: EmptyMessage,
        ) -> Transition[StoppingGoToLockdown.StoppingGoToLockdown]:
            state_machine.logger.warning(
                "Cancelling current mission due to robot going to lockdown"
            )
            state_machine.events.state_machine_events.stop_mission.trigger_event(
                EmptyMessage()
            )
            return StoppingGoToLockdown.transition(mission_id)

        def _mission_success_event_handler(
            success: EmptyMessage,
        ) -> Transition[AwaitNextMission.AwaitNextMission]:
            state_machine.logger.info("Mission succeeded")
            return AwaitNextMission.transition()

        def _set_maintenance_mode_event_handler(
            should_set_maintenance_mode: EmptyMessage,
        ) -> Transition[StoppingDueToMaintenance.StoppingDueToMaintenance]:
            state_machine.logger.warning(
                "Cancelling current mission due to robot going to maintenance mode"
            )
            state_machine.events.state_machine_events.stop_mission.trigger_event(
                EmptyMessage()
            )
            return StoppingDueToMaintenance.transition(mission_id)

        def _mission_failed_event_handler(
            mission_failed: ErrorMessage,
        ) -> Transition[AwaitNextMission.AwaitNextMission]:
            state_machine.logger.warning(
                f"Mission failed because: " f"{mission_failed.error_description}"
            )
            return AwaitNextMission.transition()

        def _stop_mission_event_handler(
            stop_mission_id: str,
        ) -> Optional[Transition[Stopping.Stopping]]:
            if mission_id == stop_mission_id or stop_mission_id == "":
                return Stopping.transition_and_trigger_stop(mission_id, True)
            else:
                state_machine.events.api_requests.stop_mission.response.trigger_event(
                    ControlMissionResponse(
                        success=False, failure_reason="Mission not found"
                    )
                )
                return None

        event_handlers: List[EventHandlerMapping] = [
            EventHandlerMapping[str](
                name="stop_mission_event",
                event=events.api_requests.stop_mission.request,
                handler=_stop_mission_event_handler,
            ),
            EventHandlerMapping[EmptyMessage](
                name="pause_mission_event",
                event=events.api_requests.pause_mission.request,
                handler=_pause_mission_event_handler,
            ),
            EventHandlerMapping[ErrorMessage](
                name="mission_failed_event",
                event=events.robot_service_events.mission_failed,
                handler=_mission_failed_event_handler,
            ),
            EventHandlerMapping[EmptyMessage](
                name="mission_succeeded_event",
                event=events.robot_service_events.mission_succeeded,
                handler=_mission_success_event_handler,
            ),
            EventHandlerMapping[float](
                name="robot_battery_update_event",
                event=shared_state.robot_battery_level,
                handler=_robot_battery_level_updated_handler,
                should_not_consume=True,
            ),
            EventHandlerMapping[EmptyMessage](
                name="send_to_lockdown_event",
                event=events.api_requests.send_to_lockdown.request,
                handler=_send_to_lockdown_event_handler,
            ),
            EventHandlerMapping[EmptyMessage](
                name="set_maintenance_mode",
                event=events.api_requests.set_maintenance_mode.request,
                handler=_set_maintenance_mode_event_handler,
            ),
        ]
        super().__init__(
            state_name=States.Monitor,
            state_machine=state_machine,
            event_handler_mappings=event_handlers,
        )


def transition_and_start_mission(
    mission: Mission, should_respond_to_API_request: bool = False
) -> Transition[Monitor]:
    def _transition(state_machine: "StateMachine") -> Monitor:
        state_machine.start_mission(mission=mission)
        if should_respond_to_API_request:
            state_machine.events.api_requests.start_mission.response.trigger_event(
                MissionStartResponse(mission_started=True)
            )
        return Monitor(state_machine, mission_id=mission.id)

    return _transition


def transition_with_existing_mission(mission_id: str) -> Transition[Monitor]:
    def _transition(state_machine: "StateMachine") -> Monitor:
        return Monitor(state_machine, mission_id=mission_id)

    return _transition



import logging
from urllib.parse import urljoin

from azure.monitor.opentelemetry.exporter import (
    AzureMonitorLogExporter,
    AzureMonitorTraceExporter,
)
from fastapi import FastAPI
from opentelemetry import trace
from opentelemetry._logs import set_logger_provider
from opentelemetry.exporter.otlp.proto.http._log_exporter import (
    OTLPLogExporter as OTLPHttpLogExporter,
)
from opentelemetry.exporter.otlp.proto.http.trace_exporter import (
    OTLPSpanExporter as OTLPHttpSpanExporter,
)
from opentelemetry.instrumentation.fastapi import FastAPIInstrumentor
from opentelemetry.sdk._logs import LoggerProvider, LoggingHandler
from opentelemetry.sdk._logs.export import BatchLogRecordProcessor
from opentelemetry.sdk.resources import SERVICE_NAME, Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor

from isar.config.log import load_log_config
from isar.config.settings import settings

logging.getLogger("opentelemetry.sdk").setLevel(logging.CRITICAL)


def setup_open_telemetry(app: FastAPI) -> None:

    service_name = settings.ROBOT_NAME
    resource = Resource.create({SERVICE_NAME: service_name})

    tracer_provider = TracerProvider(resource=resource)
    log_provider = LoggerProvider(resource=resource)

    if settings.LOG_HANDLER_APPLICATION_INSIGHTS_ENABLED:
        print("[OTEL] Azure Monitor exporters enabled")
        azure_monitor_trace_exporter, azure_monitor_log_exporter = (
            get_azure_monitor_exporters()
        )

        tracer_provider.add_span_processor(
            BatchSpanProcessor(azure_monitor_trace_exporter)
        )

        log_provider.add_log_record_processor(
            BatchLogRecordProcessor(azure_monitor_log_exporter)
        )

    otlp_exporter_endpoint = settings.OPEN_TELEMETRY_OTLP_EXPORTER_ENDPOINT
    if otlp_exporter_endpoint:
        print(f"[OTEL] OTLP exporters enabled, endpoint={otlp_exporter_endpoint}")
        otlp_trace_exporter, otlp_log_exporter = get_otlp_exporters(
            otlp_exporter_endpoint
        )
        tracer_provider.add_span_processor(BatchSpanProcessor(otlp_trace_exporter))

        log_provider.add_log_record_processor(
            BatchLogRecordProcessor(otlp_log_exporter)
        )

    set_logger_provider(log_provider)
    trace.set_tracer_provider(tracer_provider)

    handler = LoggingHandler(logger_provider=log_provider)
    attach_loggers_for_open_telemetry(handler)

    FastAPIInstrumentor.instrument_app(app, tracer_provider=tracer_provider)


def attach_loggers_for_open_telemetry(handler: LoggingHandler) -> None:
    log_config = load_log_config()

    for logger_name in log_config["loggers"].keys():
        logger = logging.getLogger(logger_name)
        logger.addHandler(handler)


def get_azure_monitor_exporters() -> (
    tuple[AzureMonitorTraceExporter, AzureMonitorLogExporter]
):
    connection_string = settings.APPLICATIONINSIGHTS_CONNECTION_STRING
    trace_exporter = AzureMonitorTraceExporter(connection_string=connection_string)
    log_exporter = AzureMonitorLogExporter(connection_string=connection_string)

    return trace_exporter, log_exporter


def get_otlp_exporters(
    endpoint: str,
) -> tuple[OTLPHttpSpanExporter, OTLPHttpLogExporter]:
    base = endpoint.rstrip("/") + "/"
    trace_ep = urljoin(base, "v1/traces")
    log_ep = urljoin(base, "v1/logs")

    print("[OTEL] Using HTTP/Protobuf protocol for OpenTelemetry export")
    print(f"[OTEL]  traces → {trace_ep}")
    print(f"[OTEL]  logs   → {log_ep}")

    return OTLPHttpSpanExporter(endpoint=trace_ep), OTLPHttpLogExporter(endpoint=log_ep)



import json
from datetime import datetime, timezone
from typing import Optional

from isar.config.settings import settings
from isar.models.status import IsarStatus
from isar.services.service_connections.mqtt.mqtt_client import props_expiry
from robot_interface.models.exceptions.robot_exceptions import ErrorMessage
from robot_interface.models.mission.status import MissionStatus
from robot_interface.models.mission.task import TASKS
from robot_interface.telemetry.mqtt_client import MqttClientInterface
from robot_interface.telemetry.payloads import (
    IsarStatusPayload,
    MissionPayload,
    TaskPayload,
)
from robot_interface.utilities.json_service import EnhancedJSONEncoder


def publish_task_status(
    mqtt_publisher: MqttClientInterface, task: TASKS, mission_id: Optional[str]
) -> None:
    """Publishes the task status to the MQTT Broker"""

    error_message: Optional[ErrorMessage] = None
    if task:
        if task.error_message:
            error_message = task.error_message

    payload: TaskPayload = TaskPayload(
        isar_id=settings.ISAR_ID,
        robot_name=settings.ROBOT_NAME,
        mission_id=mission_id,
        task_id=task.id if task else None,
        status=task.status if task else None,
        task_type=task.type if task else None,
        error_reason=error_message.error_reason if error_message else None,
        error_description=(error_message.error_description if error_message else None),
        timestamp=datetime.now(timezone.utc),
    )

    mqtt_publisher.publish(
        topic=settings.TOPIC_ISAR_TASK + f"/{task.id}",
        payload=json.dumps(payload, cls=EnhancedJSONEncoder),
        qos=1,
        retain=True,
        properties=props_expiry(settings.MQTT_MISSION_AND_TASK_EXPIRY),
    )


def publish_mission_status(
    mqtt_publisher: MqttClientInterface,
    mission_id: str,
    mission_status: MissionStatus,
    error_message: Optional[ErrorMessage],
) -> None:
    if not mqtt_publisher:
        return

    payload: MissionPayload = MissionPayload(
        isar_id=settings.ISAR_ID,
        robot_name=settings.ROBOT_NAME,
        mission_id=mission_id,
        status=mission_status,
        error_reason=error_message.error_reason if error_message else None,
        error_description=(error_message.error_description if error_message else None),
        timestamp=datetime.now(timezone.utc),
    )

    mqtt_publisher.publish(
        topic=settings.TOPIC_ISAR_MISSION + f"/{mission_id}",
        payload=json.dumps(payload, cls=EnhancedJSONEncoder),
        qos=1,
        retain=True,
        properties=props_expiry(settings.MQTT_MISSION_AND_TASK_EXPIRY),
    )


def publish_isar_status(
    mqtt_publisher: MqttClientInterface, status: IsarStatus
) -> None:
    payload: IsarStatusPayload = IsarStatusPayload(
        isar_id=settings.ISAR_ID,
        robot_name=settings.ROBOT_NAME,
        status=status,
        timestamp=datetime.now(timezone.utc),
    )

    mqtt_publisher.publish(
        topic=settings.TOPIC_ISAR_STATUS,
        payload=json.dumps(payload, cls=EnhancedJSONEncoder),
        qos=1,
        retain=True,
    )



from pytest_mock import MockerFixture

from isar.models.events import EmptyMessage
from isar.robot.robot import Robot
from isar.robot.robot_monitor_mission import RobotMonitorMissionThread
from isar.robot.robot_pause_mission import RobotPauseMissionThread
from isar.robot.robot_resume_mission import RobotResumeMissionThread
from isar.robot.robot_start_mission import RobotStartMissionThread
from isar.robot.robot_stop_mission import RobotStopMissionThread
from robot_interface.models.exceptions.robot_exceptions import ErrorMessage, ErrorReason
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.task import TakeImage, Task
from tests.test_mocks.pose import DummyPose


def test_mission_fails_to_schedule(
    mocked_robot_service: Robot, mocker: MockerFixture
) -> None:
    r_service = mocked_robot_service
    mocker.patch.object(RobotStartMissionThread, "is_alive", return_value=False)
    mock_publish_mission_status = mocker.patch(
        "isar.robot.robot.publish_mission_status"
    )

    task_1: Task = TakeImage(
        target=DummyPose.default_pose().position, robot_pose=DummyPose.default_pose()
    )
    mission: Mission = Mission(name="Dummy misson", tasks=[task_1])

    r_service.start_mission_thread = RobotStartMissionThread(
        r_service.robot, r_service.signal_thread_quitting, mission
    )
    r_service.start_mission_thread.error_message = ErrorMessage(
        error_reason=ErrorReason.RobotUnknownErrorException, error_description="test"
    )

    r_service._start_mission_done_handler()

    assert r_service.robot_service_events.mission_failed.has_event()
    mission_failed_event = r_service.robot_service_events.mission_failed.get()
    assert mission_failed_event is not None
    assert mission_failed_event.error_reason == ErrorReason.RobotUnknownErrorException

    assert r_service.monitor_mission_thread is None

    mock_publish_mission_status.assert_called_once()


def test_mission_succeeds_to_schedule(
    mocked_robot_service: Robot, mocker: MockerFixture
) -> None:
    r_service = mocked_robot_service
    mocker.patch.object(RobotStartMissionThread, "is_alive", return_value=False)
    mock_publish_mission_status = mocker.patch(
        "isar.robot.robot.publish_mission_status"
    )
    mock_monitor_mission_start = mocker.patch(
        "isar.robot.robot.RobotMonitorMissionThread.start"
    )

    task_1: Task = TakeImage(
        target=DummyPose.default_pose().position, robot_pose=DummyPose.default_pose()
    )
    mission: Mission = Mission(name="Dummy misson", tasks=[task_1])

    r_service.signal_mission_stopped.set()  # We want to test that this is cleared

    r_service.start_mission_thread = RobotStartMissionThread(
        r_service.robot, r_service.signal_thread_quitting, mission
    )

    r_service._start_mission_done_handler()

    assert not r_service.signal_mission_stopped.is_set()
    assert r_service.monitor_mission_thread is not None
    assert r_service.monitor_mission_thread.mission_id == mission.id

    mock_publish_mission_status.assert_not_called()
    mock_monitor_mission_start.assert_called_once()


def test_mission_fails_to_stop(
    mocked_robot_service: Robot, mocker: MockerFixture
) -> None:
    r_service = mocked_robot_service
    mocker.patch.object(RobotStopMissionThread, "is_alive", return_value=False)
    mocker.patch.object(RobotMonitorMissionThread, "is_alive", return_value=True)

    task_1: Task = TakeImage(
        target=DummyPose.default_pose().position, robot_pose=DummyPose.default_pose()
    )
    mission: Mission = Mission(name="Dummy misson", tasks=[task_1])

    r_service.monitor_mission_thread = RobotMonitorMissionThread(
        lambda task: None,
        r_service.robot,
        r_service.mqtt_publisher,
        r_service.signal_thread_quitting,
        r_service.signal_mission_stopped,
        mission,
    )

    r_service.stop_mission_thread = RobotStopMissionThread(
        r_service.robot, r_service.signal_thread_quitting
    )
    r_service.stop_mission_thread.error_message = ErrorMessage(
        error_reason=ErrorReason.RobotUnknownErrorException, error_description="test"
    )

    r_service._stop_mission_done_handler()

    assert r_service.robot_service_events.mission_failed_to_stop.has_event()
    mission_failed_to_stop_event = (
        r_service.robot_service_events.mission_failed_to_stop.get()
    )
    assert mission_failed_to_stop_event is not None
    assert (
        mission_failed_to_stop_event.error_reason
        == ErrorReason.RobotUnknownErrorException
    )

    assert not r_service.signal_mission_stopped.is_set()
    assert r_service.monitor_mission_thread is not None

    assert not r_service.robot_service_events.mission_successfully_stopped.has_event()

    assert not r_service.signal_mission_stopped.is_set()


def test_stop_mission_waits_for_monitor_mission(
    mocked_robot_service: Robot, mocker: MockerFixture
) -> None:
    r_service = mocked_robot_service
    mocker.patch.object(RobotStopMissionThread, "is_alive", return_value=False)
    mocker.patch.object(RobotMonitorMissionThread, "is_alive", return_value=True)

    task_1: Task = TakeImage(
        target=DummyPose.default_pose().position, robot_pose=DummyPose.default_pose()
    )
    mission: Mission = Mission(name="Dummy misson", tasks=[task_1])

    r_service.monitor_mission_thread = RobotMonitorMissionThread(
        lambda task: None,
        r_service.robot,
        r_service.mqtt_publisher,
        r_service.signal_thread_quitting,
        r_service.signal_mission_stopped,
        mission,
    )

    r_service.stop_mission_thread = RobotStopMissionThread(
        r_service.robot, r_service.signal_thread_quitting
    )

    r_service._stop_mission_done_handler()

    assert not r_service.robot_service_events.mission_successfully_stopped.has_event()
    assert r_service.signal_mission_stopped.is_set()

    mocker.patch.object(RobotMonitorMissionThread, "is_alive", return_value=False)
    r_service._stop_mission_done_handler()

    assert r_service.robot_service_events.mission_successfully_stopped.has_event()
    assert not r_service.robot_service_events.mission_failed_to_stop.has_event()
    assert not r_service.robot_service_events.mission_failed.has_event()
    assert not r_service.robot_service_events.mission_succeeded.has_event()


def test_mission_succeeds_to_stop(
    mocked_robot_service: Robot, mocker: MockerFixture
) -> None:
    r_service = mocked_robot_service
    mocker.patch.object(RobotStopMissionThread, "is_alive", return_value=False)

    r_service.stop_mission_thread = RobotStopMissionThread(
        r_service.robot, r_service.signal_thread_quitting
    )

    r_service._stop_mission_done_handler()

    assert not r_service.robot_service_events.mission_failed_to_stop.has_event()
    assert r_service.robot_service_events.mission_successfully_stopped.has_event()

    assert not r_service.robot_service_events.mission_succeeded.has_event()
    assert not r_service.robot_service_events.mission_failed.has_event()

    assert r_service.monitor_mission_thread is None


def test_monitor_mission_handler_waits_for_thread(
    mocked_robot_service: Robot, mocker: MockerFixture
) -> None:
    r_service = mocked_robot_service
    mocker.patch.object(RobotMonitorMissionThread, "is_alive", return_value=True)

    task_1: Task = TakeImage(
        target=DummyPose.default_pose().position, robot_pose=DummyPose.default_pose()
    )
    mission: Mission = Mission(name="Dummy misson", tasks=[task_1])

    r_service.monitor_mission_thread = RobotMonitorMissionThread(
        lambda task: None,
        r_service.robot,
        r_service.mqtt_publisher,
        r_service.signal_thread_quitting,
        r_service.signal_mission_stopped,
        mission,
    )

    r_service.signal_mission_stopped.set()

    r_service._monitor_mission_done_handler()

    assert r_service.monitor_mission_thread is not None


def test_monitor_mission_reports_nothing_after_mission_stopped(
    mocked_robot_service: Robot, mocker: MockerFixture
) -> None:
    r_service = mocked_robot_service
    mocker.patch.object(RobotMonitorMissionThread, "is_alive", return_value=False)

    task_1: Task = TakeImage(
        target=DummyPose.default_pose().position, robot_pose=DummyPose.default_pose()
    )
    mission: Mission = Mission(name="Dummy misson", tasks=[task_1])

    r_service.monitor_mission_thread = RobotMonitorMissionThread(
        lambda task: None,
        r_service.robot,
        r_service.mqtt_publisher,
        r_service.signal_thread_quitting,
        r_service.signal_mission_stopped,
        mission,
    )

    r_service.signal_mission_stopped.set()

    r_service._monitor_mission_done_handler()

    assert not r_service.robot_service_events.mission_succeeded.has_event()
    assert not r_service.robot_service_events.mission_failed.has_event()

    assert r_service.monitor_mission_thread is None


def test_monitor_mission_reports_mission_failed(
    mocked_robot_service: Robot, mocker: MockerFixture
) -> None:
    r_service = mocked_robot_service
    mocker.patch.object(RobotMonitorMissionThread, "is_alive", return_value=False)

    task_1: Task = TakeImage(
        target=DummyPose.default_pose().position, robot_pose=DummyPose.default_pose()
    )
    mission: Mission = Mission(name="Dummy misson", tasks=[task_1])

    r_service.monitor_mission_thread = RobotMonitorMissionThread(
        lambda task: None,
        r_service.robot,
        r_service.mqtt_publisher,
        r_service.signal_thread_quitting,
        r_service.signal_mission_stopped,
        mission,
    )

    r_service.monitor_mission_thread.error_message = ErrorMessage(
        ErrorReason.RobotUnknownErrorException, ""
    )

    r_service._monitor_mission_done_handler()

    assert not r_service.robot_service_events.mission_succeeded.has_event()
    assert r_service.robot_service_events.mission_failed.has_event()

    assert r_service.monitor_mission_thread is None


def test_monitor_mission_reports_mission_success(
    mocked_robot_service: Robot, mocker: MockerFixture
) -> None:
    r_service = mocked_robot_service
    mocker.patch.object(RobotMonitorMissionThread, "is_alive", return_value=False)

    task_1: Task = TakeImage(
        target=DummyPose.default_pose().position, robot_pose=DummyPose.default_pose()
    )
    mission: Mission = Mission(name="Dummy misson", tasks=[task_1])

    r_service.monitor_mission_thread = RobotMonitorMissionThread(
        lambda task: None,
        r_service.robot,
        r_service.mqtt_publisher,
        r_service.signal_thread_quitting,
        r_service.signal_mission_stopped,
        mission,
    )

    r_service._monitor_mission_done_handler()

    assert r_service.robot_service_events.mission_succeeded.has_event()
    assert not r_service.robot_service_events.mission_failed.has_event()

    assert r_service.monitor_mission_thread is None


def test_mission_fails_to_pause(
    mocked_robot_service: Robot, mocker: MockerFixture
) -> None:
    r_service = mocked_robot_service
    mocker.patch.object(RobotPauseMissionThread, "is_alive", return_value=False)

    r_service.pause_mission_thread = RobotPauseMissionThread(
        r_service.robot, r_service.signal_thread_quitting
    )
    r_service.pause_mission_thread.error_message = ErrorMessage(
        error_reason=ErrorReason.RobotUnknownErrorException, error_description="test"
    )

    r_service._pause_mission_done_handler()

    assert r_service.robot_service_events.mission_failed_to_pause.has_event()
    mission_failed_to_pause_event = (
        r_service.robot_service_events.mission_failed_to_pause.get()
    )
    assert mission_failed_to_pause_event is not None
    assert (
        mission_failed_to_pause_event.error_reason
        == ErrorReason.RobotUnknownErrorException
    )

    assert not r_service.robot_service_events.mission_successfully_paused.has_event()


def test_mission_succeeds_to_pause(
    mocked_robot_service: Robot, mocker: MockerFixture
) -> None:
    r_service = mocked_robot_service
    mocker.patch.object(RobotPauseMissionThread, "is_alive", return_value=False)

    r_service.pause_mission_thread = RobotPauseMissionThread(
        r_service.robot, r_service.signal_thread_quitting
    )

    r_service._pause_mission_done_handler()

    assert not r_service.robot_service_events.mission_failed_to_pause.has_event()

    assert r_service.robot_service_events.mission_successfully_paused.has_event()


def test_mission_fails_to_resume(
    mocked_robot_service: Robot, mocker: MockerFixture
) -> None:
    r_service = mocked_robot_service
    mocker.patch.object(RobotResumeMissionThread, "is_alive", return_value=False)

    r_service.resume_mission_thread = RobotResumeMissionThread(
        r_service.robot, r_service.signal_thread_quitting
    )
    r_service.resume_mission_thread.error_message = ErrorMessage(
        error_reason=ErrorReason.RobotUnknownErrorException, error_description="test"
    )

    r_service._resume_mission_done_handler()

    assert r_service.robot_service_events.mission_failed_to_resume.has_event()
    mission_failed_to_resume_event = (
        r_service.robot_service_events.mission_failed_to_resume.get()
    )
    assert mission_failed_to_resume_event is not None
    assert (
        mission_failed_to_resume_event.error_reason
        == ErrorReason.RobotUnknownErrorException
    )

    assert not r_service.robot_service_events.mission_successfully_resumed.has_event()


def test_mission_succeeds_to_resume(
    mocked_robot_service: Robot, mocker: MockerFixture
) -> None:
    r_service = mocked_robot_service
    mocker.patch.object(RobotResumeMissionThread, "is_alive", return_value=False)

    r_service.resume_mission_thread = RobotResumeMissionThread(
        r_service.robot, r_service.signal_thread_quitting
    )

    r_service._resume_mission_done_handler()

    assert not r_service.robot_service_events.mission_failed_to_resume.has_event()

    assert r_service.robot_service_events.mission_successfully_resumed.has_event()


def test_mission_stop_waits_for_mission_to_start(
    mocked_robot_service: Robot, mocker: MockerFixture
) -> None:
    r_service = mocked_robot_service
    mocker.patch.object(RobotStartMissionThread, "is_alive", return_value=True)

    task_1: Task = TakeImage(
        target=DummyPose.default_pose().position, robot_pose=DummyPose.default_pose()
    )
    mission: Mission = Mission(name="Dummy misson", tasks=[task_1])
    r_service.start_mission_thread = RobotStartMissionThread(
        r_service.robot, r_service.signal_thread_quitting, mission
    )
    r_service.stop_mission_thread = None

    r_service.state_machine_events.stop_mission.trigger_event(EmptyMessage())

    r_service._stop_mission_request_handler(r_service.state_machine_events.stop_mission)

    assert r_service.state_machine_events.stop_mission.has_event()
    assert r_service.stop_mission_thread is None


def test_mission_stop_starts_when_start_is_done(
    mocked_robot_service: Robot, mocker: MockerFixture
) -> None:
    r_service = mocked_robot_service
    mocker.patch.object(RobotStartMissionThread, "is_alive", return_value=False)

    mock_stop_mission_start = mocker.patch(
        "isar.robot.robot.RobotStopMissionThread.start"
    )

    task_1: Task = TakeImage(
        target=DummyPose.default_pose().position, robot_pose=DummyPose.default_pose()
    )
    mission: Mission = Mission(name="Dummy misson", tasks=[task_1])
    r_service.start_mission_thread = RobotStartMissionThread(
        r_service.robot, r_service.signal_thread_quitting, mission
    )
    r_service.stop_mission_thread = None

    r_service.state_machine_events.stop_mission.trigger_event(EmptyMessage())

    r_service._stop_mission_request_handler(r_service.state_machine_events.stop_mission)

    assert not r_service.state_machine_events.stop_mission.has_event()
    assert r_service.stop_mission_thread is not None
    mock_stop_mission_start.assert_called_once()


def test_mission_pause_waits_for_mission_to_start(
    mocked_robot_service: Robot, mocker: MockerFixture
) -> None:
    r_service = mocked_robot_service
    mocker.patch.object(RobotStartMissionThread, "is_alive", return_value=True)

    task_1: Task = TakeImage(
        target=DummyPose.default_pose().position, robot_pose=DummyPose.default_pose()
    )
    mission: Mission = Mission(name="Dummy misson", tasks=[task_1])
    r_service.start_mission_thread = RobotStartMissionThread(
        r_service.robot, r_service.signal_thread_quitting, mission
    )
    r_service.pause_mission_thread = None

    r_service.state_machine_events.pause_mission.trigger_event(EmptyMessage())

    r_service._pause_mission_request_handler(
        r_service.state_machine_events.pause_mission
    )

    assert r_service.state_machine_events.pause_mission.has_event()
    assert r_service.pause_mission_thread is None


def test_mission_paus_starts_when_start_is_done(
    mocked_robot_service: Robot, mocker: MockerFixture
) -> None:
    r_service = mocked_robot_service
    mocker.patch.object(RobotStartMissionThread, "is_alive", return_value=False)

    mock_pause_mission_start = mocker.patch(
        "isar.robot.robot.RobotPauseMissionThread.start"
    )

    task_1: Task = TakeImage(
        target=DummyPose.default_pose().position, robot_pose=DummyPose.default_pose()
    )
    mission: Mission = Mission(name="Dummy misson", tasks=[task_1])
    r_service.start_mission_thread = RobotStartMissionThread(
        r_service.robot, r_service.signal_thread_quitting, mission
    )
    r_service.pause_mission_thread = None

    r_service.state_machine_events.pause_mission.trigger_event(EmptyMessage())

    r_service._pause_mission_request_handler(
        r_service.state_machine_events.pause_mission
    )

    assert not r_service.state_machine_events.pause_mission.has_event()
    assert r_service.pause_mission_thread is not None
    mock_pause_mission_start.assert_called_once()



import json
import time
from abc import ABCMeta, abstractmethod
from datetime import datetime, timezone
from queue import Queue
from typing import Callable, Tuple

from paho.mqtt.packettypes import PacketTypes
from paho.mqtt.properties import Properties

from isar.config.settings import settings
from robot_interface.models.exceptions.robot_exceptions import (
    RobotTelemetryException,
    RobotTelemetryNoUpdateException,
    RobotTelemetryPoseException,
)
from robot_interface.telemetry.payloads import CloudHealthPayload
from robot_interface.utilities.json_service import EnhancedJSONEncoder


def props_expiry(seconds: int) -> Properties:
    p = Properties(PacketTypes.PUBLISH)
    p.MessageExpiryInterval = seconds
    return p


class MqttClientInterface(metaclass=ABCMeta):
    @abstractmethod
    def publish(
        self,
        topic: str,
        payload: str,
        qos: int = 0,
        retain: bool = False,
        properties: Properties = None,
    ) -> None:
        """
        Parameters
        ----------
        topic : string
            MQTT topic to publish to
        payload : string
            Payload to send to publish on the topic
        qos : integer
            Quality of Service
        retain : boolean
            Retain on topic

        Returns
        -------
        """
        pass


class MqttPublisher(MqttClientInterface):
    def __init__(self, mqtt_queue: Queue) -> None:
        self.mqtt_queue: Queue = mqtt_queue

    def publish(
        self,
        topic: str,
        payload: str,
        qos: int = 0,
        retain: bool = False,
        properties: Properties = None,
    ) -> None:
        queue_message: Tuple[str, str, int, bool, Properties] = (
            topic,
            payload,
            qos,
            retain,
            properties,
        )
        self.mqtt_queue.put(queue_message)


class MqttTelemetryPublisher(MqttClientInterface):
    def __init__(
        self,
        mqtt_queue: Queue,
        telemetry_method: Callable,
        topic: str,
        interval: float,
        qos: int = 0,
        retain: bool = False,
        properties: Properties = None,
    ) -> None:
        self.mqtt_queue: Queue = mqtt_queue
        self.telemetry_method: Callable = telemetry_method
        self.topic: str = topic
        self.interval: float = interval
        self.qos: int = qos
        self.retain: bool = retain
        self.properties: Properties = properties

    def run(self, isar_id: str, robot_name: str) -> None:
        self.cloud_health_topic: str = f"isar/{isar_id}/cloud_health"
        self.battery_topic: str = f"isar/{isar_id}/battery"
        self.pose_topic: str = f"isar/{isar_id}/pose"
        self.pressure_topic: str = f"isar/{isar_id}/pressure"
        topic: str
        payload: str

        while True:
            try:
                payload = self.telemetry_method(isar_id=isar_id, robot_name=robot_name)
                topic = self.topic
            except (RobotTelemetryPoseException, RobotTelemetryNoUpdateException):
                time.sleep(self.interval)
                continue
            except RobotTelemetryException:
                payload = json.dumps(
                    CloudHealthPayload(isar_id, robot_name, datetime.now(timezone.utc)),
                    cls=EnhancedJSONEncoder,
                )
                topic = self.cloud_health_topic

            publish_properties = self.properties

            if topic in (
                self.battery_topic,
                self.pose_topic,
                self.pressure_topic,
            ):
                publish_properties = props_expiry(settings.MQTT_TELEMETRY_EXPIRY)

            self.publish(
                topic=topic,
                payload=payload,
                qos=self.qos,
                retain=self.retain,
                properties=publish_properties,
            )
            time.sleep(self.interval)

    def publish(
        self,
        topic: str,
        payload: str,
        qos: int = 0,
        retain: bool = False,
        properties: Properties = None,
    ) -> None:
        queue_message: Tuple[str, str, int, bool, Properties] = (
            topic,
            payload,
            qos,
            retain,
            properties,
        )
        self.mqtt_queue.put(queue_message)



from typing import TYPE_CHECKING, List

import isar.state_machine.states.monitor as Monitor
import isar.state_machine.states.returning_home as ReturningHome
from isar.apis.models.models import MissionStartResponse
from isar.eventhandlers.eventhandler import EventHandlerMapping, State, Transition
from isar.models.events import EmptyMessage
from isar.state_machine.states_enum import States
from robot_interface.models.exceptions.robot_exceptions import ErrorMessage
from robot_interface.models.mission.mission import Mission

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class StoppingReturnHome(State):

    def __init__(self, state_machine: "StateMachine", mission: Mission):
        events = state_machine.events

        response = MissionStartResponse(
            mission_id=mission.id,
            mission_started=True,
        )
        state_machine.events.api_requests.start_mission.response.trigger_event(response)

        def _failed_stop_return_home_event_handler(
            error_message: ErrorMessage,
        ) -> Transition[ReturningHome.ReturningHome]:
            state_machine.logger.warning(
                f"Failed to stop return home mission {error_message.error_description}"
            )
            return ReturningHome.transition_to_existing_mission()

        event_handlers: List[EventHandlerMapping] = [
            EventHandlerMapping[ErrorMessage](
                name="failed_stop_event",
                event=events.robot_service_events.mission_failed_to_stop,
                handler=_failed_stop_return_home_event_handler,
            ),
            EventHandlerMapping[EmptyMessage](
                name="successful_stop_event",
                event=events.robot_service_events.mission_successfully_stopped,
                handler=lambda event: Monitor.transition_and_start_mission(
                    mission, True
                ),
            ),
        ]
        super().__init__(
            state_name=States.StoppingReturnHome,
            state_machine=state_machine,
            event_handler_mappings=event_handlers,
        )


def transition(mission: Mission) -> Transition[StoppingReturnHome]:
    def _transition(state_machine: "StateMachine") -> StoppingReturnHome:
        return StoppingReturnHome(state_machine, mission)

    return _transition



from typing import Optional, cast

from isar.eventhandlers.eventhandler import EventHandlerMapping, State
from isar.models.events import EmptyMessage
from isar.state_machine.state_machine import StateMachine
from isar.state_machine.states.home import Home
from isar.state_machine.states.intervention_needed import InterventionNeeded
from isar.state_machine.states.lockdown import Lockdown
from isar.state_machine.states.recharging import Recharging
from isar.state_machine.states.returning_home import ReturningHome
from robot_interface.models.exceptions.robot_exceptions import ErrorMessage, ErrorReason
from robot_interface.models.mission.status import RobotStatus


def test_lockdown_transitions_to_home(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.shared_state.robot_battery_level.trigger_event(80.0)
    sync_state_machine.current_state = Lockdown(sync_state_machine)

    lockdown_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        lockdown_state.get_event_handler_by_name("release_from_lockdown")
    )

    assert event_handler is not None

    transition = event_handler.handler(EmptyMessage())

    assert sync_state_machine.events.api_requests.release_from_lockdown.response.check()
    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is Home


def test_state_machine_with_return_home_failure_successful_retries(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.shared_state.robot_battery_level.trigger_event(80.0)
    sync_state_machine.current_state = ReturningHome(sync_state_machine)

    returning_home_state: State = cast(State, sync_state_machine.current_state)
    event_handler_success: Optional[EventHandlerMapping] = (
        returning_home_state.get_event_handler_by_name("mission_succeeded_event")
    )
    event_handler_failure: Optional[EventHandlerMapping] = (
        returning_home_state.get_event_handler_by_name("mission_failed_event")
    )

    # We do not retry return home missions if the robot is not ready for another mission
    sync_state_machine.shared_state.robot_status.trigger_event(RobotStatus.Available)

    assert event_handler_success is not None
    assert event_handler_failure is not None

    transition = event_handler_failure.handler(
        ErrorMessage(
            error_reason=ErrorReason.RobotUnknownErrorException,
            error_description="test",
        )
    )

    assert transition is None  # type: ignore
    assert sync_state_machine.current_state.failed_return_home_attempts == 1

    transition = event_handler_success.handler(EmptyMessage())

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is Home


def test_intervention_needed_transitions_to_home_if_robot_is_home(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.current_state = InterventionNeeded(sync_state_machine)

    intervention_needed_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        intervention_needed_state.get_event_handler_by_name("robot_status_event")
    )
    assert event_handler is not None

    sync_state_machine.shared_state.robot_status.trigger_event(RobotStatus.Home)

    transition = event_handler.handler(EmptyMessage())

    assert transition is not None

    sync_state_machine.current_state = transition(sync_state_machine)
    assert type(sync_state_machine.current_state) is Home


def test_recharging_goes_to_home_when_battery_high(
    sync_state_machine: StateMachine,
) -> None:
    sync_state_machine.current_state = Recharging(sync_state_machine)

    recharging_state: State = cast(State, sync_state_machine.current_state)
    event_handler: Optional[EventHandlerMapping] = (
        recharging_state.get_event_handler_by_name("robot_battery_update_event")
    )

    assert event_handler is not None

    transition = event_handler.handler(99.9)

    sync_state_machine.current_state = transition(sync_state_machine)

    assert type(sync_state_machine.current_state) is Home



from typing import TYPE_CHECKING, List, Optional, Union

import isar.state_machine.states.await_next_mission as AwaitNextMission
import isar.state_machine.states.blocked_protective_stop as BlockedProtectiveStop
import isar.state_machine.states.home as Home
import isar.state_machine.states.maintenance as Maintenance
import isar.state_machine.states.offline as Offline
import isar.state_machine.states.stopping as Stopping
from isar.apis.models.models import MaintenanceResponse
from isar.eventhandlers.eventhandler import EventHandlerMapping, State, Transition
from isar.models.events import EmptyMessage
from isar.state_machine.states_enum import States
from robot_interface.models.mission.status import RobotStatus

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class UnknownStatus(State):

    def __init__(self, state_machine: "StateMachine"):
        # Ensures that we will check the status immediately instead of waiting for it to change
        events = state_machine.events
        shared_state = state_machine.shared_state
        events.robot_service_events.robot_status_changed.trigger_event(EmptyMessage())

        def _set_maintenance_mode_event_handler(
            should_set_maintenance_mode: EmptyMessage,
        ) -> Transition[Maintenance.Maintenance]:
            events.api_requests.set_maintenance_mode.response.trigger_event(
                MaintenanceResponse(is_maintenance_mode=True)
            )
            return Maintenance.transition()

        def _robot_status_event_handler(
            has_changed: EmptyMessage,
        ) -> Optional[
            Union[
                Transition[Home.Home],
                Transition[AwaitNextMission.AwaitNextMission],
                Transition[Offline.Offline],
                Transition[BlockedProtectiveStop.BlockedProtectiveStop],
                Transition[Stopping.Stopping],
            ]
        ]:
            robot_status: Optional[RobotStatus] = shared_state.robot_status.check()

            if robot_status == RobotStatus.Home:
                self.logger.info(
                    "Got robot status home while in unknown status state. Leaving unknown status state."
                )
                return Home.transition()
            elif robot_status == RobotStatus.Available:
                self.logger.info(
                    "Got robot status available while in unknown status state. Leaving unknown status state."
                )
                return AwaitNextMission.transition()
            elif robot_status == RobotStatus.Offline:
                self.logger.info(
                    "Got robot status offline while in unknown status state. Leaving unknown status state."
                )
                return Offline.transition()
            elif robot_status == RobotStatus.BlockedProtectiveStop:
                self.logger.info(
                    "Got robot status blocked protective stop while in unknown status state. Leaving unknown status state."
                )
                return BlockedProtectiveStop.transition()
            elif robot_status == RobotStatus.Busy:
                self.logger.info(
                    "Got robot status busy while in unknown status state. Leaving unknown status state."
                )
                return Stopping.transition_and_trigger_stop("")
            return None

        def _stop_mission_event_handler(
            mission_id: str,
        ) -> Transition[Stopping.Stopping]:
            return Stopping.transition_and_trigger_stop(mission_id, True)

        event_handlers: List[EventHandlerMapping] = [
            EventHandlerMapping[str](
                name="stop_mission_event",
                event=events.api_requests.stop_mission.request,
                handler=_stop_mission_event_handler,
            ),
            EventHandlerMapping[EmptyMessage](
                name="robot_status_event",
                event=events.robot_service_events.robot_status_changed,
                handler=_robot_status_event_handler,
            ),
            EventHandlerMapping[EmptyMessage](
                name="set_maintenance_mode",
                event=events.api_requests.set_maintenance_mode.request,
                handler=_set_maintenance_mode_event_handler,
            ),
        ]
        super().__init__(
            state_name=States.UnknownStatus,
            state_machine=state_machine,
            event_handler_mappings=event_handlers,
        )


def transition() -> Transition[UnknownStatus]:
    def _transition(state_machine: "StateMachine") -> UnknownStatus:
        return UnknownStatus(state_machine)

    return _transition



from typing import TYPE_CHECKING, List, Optional

import isar.state_machine.states.going_to_lockdown as GoingToLockdown
import isar.state_machine.states.going_to_recharging as GoingToRecharging
import isar.state_machine.states.home as Home
import isar.state_machine.states.intervention_needed as InterventionNeeded
import isar.state_machine.states.pausing_return_home as PausingReturnHome
import isar.state_machine.states.stopping_due_to_maintenance as StoppingDueToMaintenance
import isar.state_machine.states.stopping_return_home as StoppingReturnHome
from isar.apis.models.models import (
    ControlMissionResponse,
    LockdownResponse,
    MissionStartResponse,
)
from isar.config.settings import settings
from isar.eventhandlers.eventhandler import EventHandlerMapping, State, Transition
from isar.models.events import EmptyMessage
from isar.state_machine.states_enum import States
from robot_interface.models.exceptions.robot_exceptions import ErrorMessage
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.task import ReturnToHome

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class ReturningHome(State):

    def __init__(self, state_machine: "StateMachine"):
        self.failed_return_home_attempts: int = 0
        events = state_machine.events
        shared_state = state_machine.shared_state

        def _pause_mission_event_handler(
            should_pause: EmptyMessage,
        ) -> Transition[PausingReturnHome.PausingReturnHome]:
            state_machine.events.api_requests.pause_mission.response.trigger_event(
                ControlMissionResponse(success=True)
            )
            state_machine.events.state_machine_events.pause_mission.trigger_event(
                EmptyMessage()
            )
            return PausingReturnHome.transition()

        def _start_mission_event_handler(
            mission: Mission,
        ) -> Optional[Transition[StoppingReturnHome.StoppingReturnHome]]:
            # The check below is arguably not needed due to the battery eventhandler
            if not state_machine.battery_level_is_above_mission_start_threshold():
                response = MissionStartResponse(
                    mission_id=None,
                    mission_started=False,
                    mission_not_started_reason="Robot battery too low",
                )
                state_machine.events.api_requests.start_mission.response.trigger_event(
                    response
                )
                return None

            state_machine.events.state_machine_events.stop_mission.trigger_event(
                EmptyMessage()
            )
            return StoppingReturnHome.transition(mission)

        def _mission_success_event_handler(
            success: EmptyMessage,
        ) -> Transition[Home.Home]:
            return Home.transition()

        def _send_to_lockdown_event_handler(
            should_lockdown: EmptyMessage,
        ) -> Transition[GoingToLockdown.GoingToLockdown]:
            events.api_requests.send_to_lockdown.response.trigger_event(
                LockdownResponse(lockdown_started=True)
            )
            return GoingToLockdown.transition()

        def _mission_failed_event_handler(
            mission_failed: ErrorMessage,
        ) -> Optional[Transition[InterventionNeeded.InterventionNeeded]]:
            state_machine.logger.warning(
                f"Failed return home because: " f"{mission_failed.error_description}"
            )
            self.failed_return_home_attempts += 1
            if self.failed_return_home_attempts >= settings.RETURN_HOME_RETRY_LIMIT:
                state_machine.logger.warning(
                    f"Failed to return home after {self.failed_return_home_attempts} attempts."
                )
                state_machine.publish_intervention_needed(
                    error_message=f"Return home failed after {self.failed_return_home_attempts} attempts."
                )
                return InterventionNeeded.transition()
            else:
                state_machine.start_mission(
                    Mission(
                        tasks=[ReturnToHome()],
                        name="Return Home",
                    )
                )
                return None

        def _set_maintenance_mode_event_handler(
            should_set_maintenance_mode: EmptyMessage,
        ) -> Transition[StoppingDueToMaintenance.StoppingDueToMaintenance]:
            state_machine.logger.warning(
                "Cancelling current mission due to robot going to maintenance mode"
            )
            state_machine.events.state_machine_events.stop_mission.trigger_event(
                EmptyMessage()
            )
            return StoppingDueToMaintenance.transition("")

        def _robot_already_home_event_handler(
            already_home: EmptyMessage,
        ) -> Transition[Home.Home]:
            state_machine.logger.info(
                "Robot reported that it is already home. "
                "Assuming return home mission successful without running."
            )
            return Home.transition()

        def _robot_battery_level_updated_handler(
            battery_level: float,
        ) -> Optional[Transition[GoingToRecharging.GoingToRecharging]]:
            if (
                battery_level is None
                or battery_level >= settings.ROBOT_MISSION_BATTERY_START_THRESHOLD
            ):
                return None

            return GoingToRecharging.transition()

        event_handlers: List[EventHandlerMapping] = [
            EventHandlerMapping[EmptyMessage](
                name="pause_mission_event",
                event=events.api_requests.pause_mission.request,
                handler=_pause_mission_event_handler,
            ),
            EventHandlerMapping[ErrorMessage](
                name="mission_failed_event",
                event=events.robot_service_events.mission_failed,
                handler=_mission_failed_event_handler,
            ),
            EventHandlerMapping[Mission](
                name="start_mission_event",
                event=events.api_requests.start_mission.request,
                handler=_start_mission_event_handler,
            ),
            EventHandlerMapping[EmptyMessage](
                name="mission_succeeded_event",
                event=events.robot_service_events.mission_succeeded,
                handler=_mission_success_event_handler,
            ),
            EventHandlerMapping[float](
                name="robot_battery_update_event",
                event=shared_state.robot_battery_level,
                handler=_robot_battery_level_updated_handler,
                should_not_consume=True,
            ),
            EventHandlerMapping[EmptyMessage](
                name="send_to_lockdown_event",
                event=events.api_requests.send_to_lockdown.request,
                handler=_send_to_lockdown_event_handler,
            ),
            EventHandlerMapping[EmptyMessage](
                name="set_maintenance_mode",
                event=events.api_requests.set_maintenance_mode.request,
                handler=_set_maintenance_mode_event_handler,
            ),
            EventHandlerMapping[EmptyMessage](
                name="robot_already_home",
                event=events.robot_service_events.robot_already_home,
                handler=_robot_already_home_event_handler,
            ),
        ]
        super().__init__(
            state_name=States.ReturningHome,
            state_machine=state_machine,
            event_handler_mappings=event_handlers,
        )


def transition_and_start_mission(
    should_respond_to_API_request: bool = False,
) -> Transition[ReturningHome]:
    def _transition(state_machine: "StateMachine") -> ReturningHome:
        state_machine.start_return_home_mission()
        if should_respond_to_API_request:
            state_machine.events.api_requests.return_home.response.trigger_event(
                EmptyMessage()
            )
        return ReturningHome(state_machine)

    return _transition


def transition_to_existing_mission() -> Transition[ReturningHome]:
    def _transition(state_machine: "StateMachine") -> ReturningHome:
        return ReturningHome(state_machine)

    return _transition



import re
from http import HTTPStatus
from unittest import mock
from uuid import uuid4

import pytest
from fastapi.encoders import jsonable_encoder
from fastapi.testclient import TestClient
from pytest_mock import MockerFixture

from isar.apis.models.models import ControlMissionResponse
from isar.apis.models.start_mission_definition import StopMissionDefinition
from isar.models.events import EventTimeoutError
from isar.services.utilities.scheduling_utilities import SchedulingUtilities
from isar.state_machine.states_enum import States
from tests.test_mocks.mission_definition import DummyMissionDefinition

dummy_mission = DummyMissionDefinition.default_mission
dummy_mission_stopped = DummyMissionDefinition.stopped_mission

mock_return_unknown_status = mock.Mock(return_value=States.UnknownStatus)
mock_return_robot_home_then_monitor = mock.Mock(
    side_effect=[States.Home, States.Monitor]
)
mock_return_monitor = mock.Mock(return_value=States.Monitor)
mock_return_paused = mock.Mock(return_value=States.Paused)
mock_return_home = mock.Mock(return_value=States.Home)
mock_void = mock.Mock()

dummy_task = dummy_mission.tasks[0]
dummy_start_mission_response = DummyMissionDefinition.dummy_start_mission_response
dummy_control_mission_response = ControlMissionResponse(success=True)
dummy_stopped_control_mission_response = ControlMissionResponse(success=True)
mock_return_control_mission_response = mock.Mock(
    return_value=dummy_control_mission_response
)
mock_return_stopped_control_mission_response = mock.Mock(
    return_value=dummy_stopped_control_mission_response
)
mock_queue_timeout_error = mock.Mock(side_effect=EventTimeoutError)

dummy_stopped_with_mission_id_control_mission_response = ControlMissionResponse(
    success=True
)
mock_return_control_mission_stop_wrong_id_response = mock.Mock(
    return_value=ControlMissionResponse(success=False, failure_reason="")
)


class TestStartMission:
    schedule_start_mission_path = "/schedule/start-mission"
    dummy_start_mission_definition = (
        DummyMissionDefinition.dummy_start_mission_definition
    )
    dummy_start_mission_content = {"mission_definition": dummy_start_mission_definition}
    dummy_start_mission_definition_image_and_thermal = {
        "mission_definition": DummyMissionDefinition.dummy_start_mission_definition_image_and_thermal
    }

    @mock.patch.object(SchedulingUtilities, "get_state", mock_return_home)
    @mock.patch.object(SchedulingUtilities, "start_mission", mock_void)
    def test_start_mission(self, client: TestClient) -> None:
        response = client.post(
            url=self.schedule_start_mission_path,
            json=jsonable_encoder(self.dummy_start_mission_content),
        )
        assert response.status_code == HTTPStatus.OK

    def test_incomplete_request(self, client: TestClient) -> None:
        response = client.post(url=self.schedule_start_mission_path, json={})
        assert response.status_code == HTTPStatus.UNPROCESSABLE_ENTITY

    @mock.patch.object(SchedulingUtilities, "get_state", mock_return_monitor)
    def test_state_machine_in_conflicting_state(self, client: TestClient) -> None:
        response = client.post(
            url=self.schedule_start_mission_path,
            json=jsonable_encoder(self.dummy_start_mission_content),
        )
        assert response.status_code == HTTPStatus.CONFLICT

    @mock.patch.object(SchedulingUtilities, "get_state", mock_return_home)
    @mock.patch.object(SchedulingUtilities, "_send_command", mock_queue_timeout_error)
    def test_start_mission_timeout(self, client: TestClient) -> None:
        response = client.post(
            url=self.schedule_start_mission_path,
            json=jsonable_encoder(self.dummy_start_mission_content),
        )
        assert response.status_code == HTTPStatus.CONFLICT
        assert response.json() == {
            "detail": "State machine has entered a state which cannot start a mission"
        }

    @mock.patch.object(SchedulingUtilities, "get_state", mock_return_home)
    @mock.patch("isar.config.settings.robot_settings.CAPABILITIES", [])
    @mock.patch.object(SchedulingUtilities, "start_mission", mock_void)
    def test_robot_not_capable(self, client: TestClient) -> None:
        response = client.post(
            url=self.schedule_start_mission_path,
            json=jsonable_encoder(
                self.dummy_start_mission_definition_image_and_thermal
            ),
        )
        assert response.status_code == HTTPStatus.BAD_REQUEST
        response_detail = response.json()["detail"]
        assert re.match(
            "Bad Request - Robot is not capable of performing mission.", response_detail
        )
        assert re.search("take_thermal_image", response_detail)
        assert re.search("take_image", response_detail)


class TestPauseMission:
    schedule_pause_mission_path = "/schedule/pause-mission"

    @mock.patch.object(SchedulingUtilities, "get_state", mock_return_monitor)
    @mock.patch.object(
        SchedulingUtilities, "_send_command", mock_return_control_mission_response
    )
    def test_pause_mission(self, client: TestClient) -> None:
        response = client.post(url=self.schedule_pause_mission_path)
        assert response.status_code == HTTPStatus.OK
        assert response.json() == jsonable_encoder(dummy_control_mission_response)

    @mock.patch.object(SchedulingUtilities, "get_state", mock_return_home)
    def test_state_machine_in_conflicting_state(self, client: TestClient) -> None:
        response = client.post(url=self.schedule_pause_mission_path)
        assert response.status_code == HTTPStatus.CONFLICT

    @mock.patch.object(SchedulingUtilities, "get_state", mock_return_monitor)
    @mock.patch.object(SchedulingUtilities, "_send_command", mock_queue_timeout_error)
    def test_pause_mission_timeout(self, client: TestClient) -> None:
        response = client.post(url=self.schedule_pause_mission_path)
        assert response.status_code == HTTPStatus.CONFLICT
        assert response.json() == {
            "detail": "State machine has entered a state which cannot pause a mission"
        }


class TestResumeMission:
    schedule_resume_mission_path = "/schedule/resume-mission"

    @mock.patch.object(SchedulingUtilities, "get_state", mock_return_paused)
    @mock.patch.object(
        SchedulingUtilities, "_send_command", mock_return_control_mission_response
    )
    def test_resume_mission(self, client: TestClient) -> None:
        response = client.post(url=self.schedule_resume_mission_path)
        assert response.status_code == HTTPStatus.OK
        assert response.json() == jsonable_encoder(dummy_control_mission_response)

    @mock.patch.object(SchedulingUtilities, "get_state", mock_return_home)
    def test_state_machine_in_conflicting_state(self, client: TestClient) -> None:
        response = client.post(url=self.schedule_resume_mission_path)
        assert response.status_code == HTTPStatus.CONFLICT

    @mock.patch.object(SchedulingUtilities, "get_state", mock_return_paused)
    @mock.patch.object(SchedulingUtilities, "_send_command", mock_queue_timeout_error)
    def test_resume_mission_timeout(self, client: TestClient) -> None:
        response = client.post(url=self.schedule_resume_mission_path)
        assert response.status_code == HTTPStatus.INTERNAL_SERVER_ERROR
        assert response.json() == {
            "detail": "Internal Server Error - Failed to resume mission"
        }


class TestStopMission:
    schedule_stop_mission_path = "/schedule/stop-mission"
    valid_states = [
        States.AwaitNextMission,
        States.Monitor,
        States.Paused,
    ]

    @pytest.mark.parametrize("state", valid_states)
    @mock.patch.object(
        SchedulingUtilities,
        "_send_command",
        mock_return_stopped_control_mission_response,
    )
    def test_stop_mission(
        self, client: TestClient, state: States, mocker: MockerFixture
    ) -> None:
        mocker.patch.object(SchedulingUtilities, "get_state", return_value=state)
        response = client.post(
            url=self.schedule_stop_mission_path,
            json=jsonable_encoder({"mission_id": StopMissionDefinition(mission_id="")}),
        )
        assert response.status_code == HTTPStatus.OK
        assert response.json() == jsonable_encoder(
            dummy_stopped_control_mission_response
        )

    @mock.patch.object(SchedulingUtilities, "get_state", mock_return_unknown_status)
    @mock.patch.object(
        SchedulingUtilities, "stop_mission", dummy_control_mission_response
    )
    def test_can_not_stop_mission_in_unknown_status(self, client: TestClient) -> None:
        response = client.post(url=self.schedule_stop_mission_path)
        assert response.status_code == HTTPStatus.CONFLICT

    @mock.patch.object(SchedulingUtilities, "get_state", mock_return_monitor)
    @mock.patch.object(SchedulingUtilities, "_send_command", mock_queue_timeout_error)
    def test_stop_mission_timeout(self, client: TestClient) -> None:
        response = client.post(
            url=self.schedule_stop_mission_path,
            json=jsonable_encoder({"mission_id": StopMissionDefinition(mission_id="")}),
        )
        assert response.status_code == HTTPStatus.CONFLICT

    @mock.patch.object(SchedulingUtilities, "get_state", mock_return_monitor)
    @mock.patch.object(
        SchedulingUtilities,
        "_send_command",
        mock_return_control_mission_stop_wrong_id_response,
    )
    def test_stop_mission_with_mission_id(self, client: TestClient) -> None:
        response = client.post(
            url=self.schedule_stop_mission_path,
            json=jsonable_encoder(
                {"mission_id": StopMissionDefinition(mission_id=str(uuid4()))}
            ),
        )
        assert response.status_code == HTTPStatus.SERVICE_UNAVAILABLE


class TestInfoRobotSettings:
    def test_info_robot_settings(self, client: TestClient) -> None:
        response = client.get(url="/info/robot-settings")
        assert response.status_code == HTTPStatus.OK



from isar.state_machine.state_machine import StateMachine
from isar.state_machine.states_enum import States


def test_initial_unknown_status(state_machine: StateMachine) -> None:
    assert state_machine.current_state.name == States.UnknownStatus



from typing import TYPE_CHECKING, List, Optional

import isar.state_machine.states.going_to_lockdown as GoingToLockdown
import isar.state_machine.states.going_to_recharging as GoingToRecharging
import isar.state_machine.states.maintenance as Maintenance
import isar.state_machine.states.monitor as Monitor
import isar.state_machine.states.returning_home as ReturningHome
import isar.state_machine.states.stopping as Stopping
from isar.apis.models.models import (
    LockdownResponse,
    MaintenanceResponse,
    MissionStartResponse,
)
from isar.config.settings import settings
from isar.eventhandlers.eventhandler import (
    EventHandlerMapping,
    State,
    TimeoutHandlerMapping,
    Transition,
)
from isar.models.events import EmptyMessage
from isar.state_machine.states_enum import States
from robot_interface.models.mission.mission import Mission

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class AwaitNextMission(State):

    def __init__(self, state_machine: "StateMachine"):
        events = state_machine.events
        shared_state = state_machine.shared_state

        def _send_to_lockdown_event_handler(
            should_lockdown: EmptyMessage,
        ) -> Transition[GoingToLockdown.GoingToLockdown]:
            events.api_requests.send_to_lockdown.response.trigger_event(
                LockdownResponse(lockdown_started=True)
            )

            state_machine.start_return_home_mission()
            return GoingToLockdown.transition()

        def _robot_battery_level_updated_handler(
            battery_level: float,
        ) -> Optional[Transition[GoingToRecharging.GoingToRecharging]]:
            if (
                battery_level is None
                or battery_level >= settings.ROBOT_MISSION_BATTERY_START_THRESHOLD
            ):
                return None

            state_machine.start_return_home_mission()
            return GoingToRecharging.transition()

        def _set_maintenance_mode_event_handler(
            should_set_maintenance_mode: EmptyMessage,
        ) -> Transition[Maintenance.Maintenance]:
            events.api_requests.set_maintenance_mode.response.trigger_event(
                MaintenanceResponse(is_maintenance_mode=True)
            )
            return Maintenance.transition()

        def _stop_mission_event_handler(
            mission_id: str,
        ) -> Transition[Stopping.Stopping]:
            return Stopping.transition_and_trigger_stop(mission_id, True)

        def _start_mission_event_handler(
            mission: Mission,
        ) -> Optional[Transition[Monitor.Monitor]]:
            if not state_machine.battery_level_is_above_mission_start_threshold():
                events.api_requests.start_mission.response.trigger_event(
                    MissionStartResponse(
                        mission_id=mission.id,
                        mission_started=False,
                        mission_not_started_reason="Robot battery too low",
                    )
                )
                return None
            return Monitor.transition_and_start_mission(mission, True)

        event_handlers: List[EventHandlerMapping] = [
            EventHandlerMapping[Mission](
                name="start_mission_event",
                event=events.api_requests.start_mission.request,
                handler=_start_mission_event_handler,
            ),
            EventHandlerMapping[EmptyMessage](
                name="return_home_event",
                event=events.api_requests.return_home.request,
                handler=lambda event: ReturningHome.transition_and_start_mission(True),
            ),
            EventHandlerMapping[str](
                name="stop_mission_event",
                event=events.api_requests.stop_mission.request,
                handler=_stop_mission_event_handler,
            ),
            EventHandlerMapping[EmptyMessage](
                name="send_to_lockdown_event",
                event=events.api_requests.send_to_lockdown.request,
                handler=_send_to_lockdown_event_handler,
            ),
            EventHandlerMapping[float](
                name="robot_battery_update_event",
                event=shared_state.robot_battery_level,
                handler=_robot_battery_level_updated_handler,
                should_not_consume=True,
            ),
            EventHandlerMapping[EmptyMessage](
                name="set_maintenance_mode",
                event=events.api_requests.set_maintenance_mode.request,
                handler=_set_maintenance_mode_event_handler,
            ),
        ]

        timers: List[TimeoutHandlerMapping] = [
            TimeoutHandlerMapping(
                name="should_return_home_timer",
                timeout_in_seconds=settings.RETURN_HOME_DELAY,
                handler=lambda: ReturningHome.transition_and_start_mission(),
            )
        ]

        super().__init__(
            state_name=States.AwaitNextMission,
            state_machine=state_machine,
            event_handler_mappings=event_handlers,
            timers=timers,
        )


def transition() -> Transition[AwaitNextMission]:
    def _transition(state_machine: "StateMachine") -> AwaitNextMission:
        return AwaitNextMission(state_machine)

    return _transition



import pytest

from isar.models.events import Event, EventTimeoutError


class TestQueueUtilities:
    @pytest.mark.parametrize(
        "message, queue_timeout, expected_message",
        [
            (
                "Test",
                3,
                "Test",
            ),
            (None, 1, None),
        ],
    )
    def test_check_queue_with_queue_size_one(
        self, message: str, queue_timeout: int, expected_message: str
    ) -> None:
        test_event: Event = Event("test")
        if message is not None:
            test_event.put(message)
            message = test_event.consume_event(timeout=queue_timeout)
            assert message == expected_message
        else:
            with pytest.raises(EventTimeoutError):
                test_event.consume_event(timeout=queue_timeout)

    def test_clear_queue(self) -> None:
        test_event: Event = Event("test")
        test_event.put(1)
        test_event.clear_event()
        assert test_event.empty()



"""Create robot state table

Revision ID: e92812b07fe8
Revises:
Create Date: 2025-11-17 12:38:45.581082

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op  # type: ignore[attr-defined]

# revision identifiers, used by Alembic.
revision: str = "e92812b07fe8"
down_revision: Union[str, Sequence[str], None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    """Upgrade schema."""
    op.create_table(
        "persistent_robot_state",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("robot_id", sa.String(length=64), nullable=False),
        sa.Column("is_maintenance_mode", sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )


def downgrade() -> None:
    """Downgrade schema."""
    op.drop_table("persistent_robot_state")



from typing import List

from isar.storage.storage_interface import (
    BlobStoragePath,
    StorageException,
    StorageInterface,
    StoragePaths,
)
from robot_interface.models.inspection.inspection import Inspection, InspectionBlob
from robot_interface.models.mission.mission import Mission


class StorageFake(StorageInterface):
    will_fail: bool = False

    def __init__(self) -> None:
        self.stored_inspections: List[Inspection] = []

    def store(self, inspection: Inspection, mission: Mission) -> None:
        if self.will_fail:
            raise StorageException("Fake failed on purpose")
        self.stored_inspections.append(inspection)

    def blob_exists(self, inspection: Inspection) -> bool:
        return inspection in self.stored_inspections


class StorageEmptyBlobPathsFake(StorageInterface):

    def __init__(self) -> None:
        self.stored: List[Inspection] = []
        self.fail: bool = False

    def store(
        self, inspection: InspectionBlob, mission: Mission
    ) -> StoragePaths[BlobStoragePath]:
        if self.fail:
            raise StorageException("fail on purpose")
        self.stored.append(inspection)
        empty = BlobStoragePath(storage_account="", blob_container="", blob_name="")
        return StoragePaths(data_path=empty, metadata_path=empty)



from typing import TYPE_CHECKING, List

import isar.state_machine.states.intervention_needed as InterventionNeeded
import isar.state_machine.states.lockdown as Lockdown
from isar.eventhandlers.eventhandler import EventHandlerMapping, State, Transition
from isar.models.events import EmptyMessage
from isar.state_machine.states_enum import States
from robot_interface.models.exceptions.robot_exceptions import ErrorMessage

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class GoingToLockdown(State):

    def __init__(self, state_machine: "StateMachine"):
        events = state_machine.events

        def _mission_failed_event_handler(
            mission_failed: ErrorMessage,
        ) -> Transition[InterventionNeeded.InterventionNeeded]:
            state_machine.logger.warning(
                f"Failed to go to lockdown because: "
                f"{mission_failed.error_description}"
            )
            state_machine.publish_intervention_needed(
                error_message="Lockdown mission failed."
            )
            return InterventionNeeded.transition()

        def _mission_failed_to_resume_event_handler(
            mission_failed_to_resume: ErrorMessage,
        ) -> Transition[InterventionNeeded.InterventionNeeded]:
            state_machine.logger.warning(
                f"Failed to resume return to home mission and going to lockdown because: "
                f"{mission_failed_to_resume.error_description or ''}"
            )
            return InterventionNeeded.transition()

        def _mission_success_event_handler(
            success: EmptyMessage,
        ) -> Transition[Lockdown.Lockdown]:
            state_machine.print_transitions()
            return Lockdown.transition()

        event_handlers: List[EventHandlerMapping] = [
            EventHandlerMapping[ErrorMessage](
                name="mission_failed_event",
                event=events.robot_service_events.mission_failed,
                handler=_mission_failed_event_handler,
            ),
            EventHandlerMapping[ErrorMessage](
                name="mission_failed_to_resume",
                event=events.robot_service_events.mission_failed_to_resume,
                handler=_mission_failed_to_resume_event_handler,
            ),
            EventHandlerMapping[EmptyMessage](
                name="mission_succeeded_event",
                event=events.robot_service_events.mission_succeeded,
                handler=_mission_success_event_handler,
            ),
        ]
        super().__init__(
            state_name=States.GoingToLockdown,
            state_machine=state_machine,
            event_handler_mappings=event_handlers,
        )


def transition() -> Transition[GoingToLockdown]:
    def _transition(state_machine: "StateMachine") -> GoingToLockdown:
        return GoingToLockdown(state_machine)

    return _transition



import logging
from copy import deepcopy
from http import HTTPStatus
from typing import List, TypeVar

from fastapi import HTTPException

from isar.apis.models.models import ControlMissionResponse, MaintenanceResponse
from isar.config.settings import settings
from isar.models.events import (
    APIEvent,
    APIRequests,
    EmptyMessage,
    EventConflictError,
    Events,
    EventTimeoutError,
    SharedState,
)
from isar.services.service_connections.persistent_memory import (
    change_persistent_robot_state_is_maintenance_mode,
)
from isar.state_machine.states_enum import States
from robot_interface.models.mission.mission import Mission

T1 = TypeVar("T1")
T2 = TypeVar("T2")


class SchedulingUtilities:
    """
    Contains utility functions for scheduling missions from the API. The class handles
    required thread communication through queues to the state machine.
    """

    def __init__(
        self,
        events: Events,
        shared_state: SharedState,
    ):
        self.api_events: APIRequests = events.api_requests
        self.shared_state: SharedState = shared_state
        self.logger = logging.getLogger("api")

    def get_state(self) -> States:
        """Return the current state of the state machine

        Raises
        ------
        HTTPException 500 Internal Server Error
            If the current state is not available on the queue
        """
        current_state = self.shared_state.state.check()
        if current_state is None:
            error_message: str = (
                "Internal Server Error - Current state of the state machine is unknown"
            )
            self.logger.error(error_message)
            raise HTTPException(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR, detail=error_message
            )
        return current_state

    def verify_robot_capable_of_mission(
        self, mission: Mission, robot_capabilities: List[str]
    ) -> bool:
        """Verify that the robot is capable of performing this mission

        Raises
        ------
        HTTPException 400 Bad request
            If the robot is not capable of performing mission
        """
        missing_capabilities = {
            task.type for task in mission.tasks if task.type not in robot_capabilities
        }

        if missing_capabilities:
            error_message = (
                f"Bad Request - Robot is not capable of performing mission."
                f" Missing functionalities: {missing_capabilities}."
            )
            self.logger.error(error_message)
            raise HTTPException(
                status_code=HTTPStatus.BAD_REQUEST,
                detail=error_message,
            )

        return True

    def verify_state_machine_ready_to_receive_mission(self, state: States) -> bool:
        """Verify that the state machine is ready to receive a mission

        Raises
        ------
        HTTPException 409 Conflict
            If state machine is not home, robot standing still, awaiting next mission
            return home paused or returning home and therefore cannot start a new mission
        """
        if (
            state == States.Home
            or state == States.AwaitNextMission
            or state == States.ReturningHome
            or state == States.ReturnHomePaused
        ):
            return True

        error_message = f"Conflict - Robot is not home, robot standing still, awaiting next mission or returning home - State: {state}"
        self.logger.warning(error_message)
        raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)

    def verify_state_machine_ready_to_receive_return_home_mission(
        self, state: States
    ) -> bool:
        """Verify that the state machine is ready to receive a return home mission

        Raises
        ------
        HTTPException 409 Conflict
            If state machine is not home, robot standing still or awaiting next mission
            and therefore cannot start a new return home mission
        """
        if state == States.Home or state == States.AwaitNextMission:
            return True

        error_message = f"Conflict - Robot is not home, robot standing still or awaiting next mission - State: {state}"
        self.logger.warning(error_message)
        raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)

    def log_mission_overview(self, mission: Mission) -> None:
        """Log an overview of the tasks in a mission"""
        log_statements: List[str] = []
        for task in mission.tasks:
            log_statements.append(
                f"{type(task).__name__:<20} {str(task.id)[:8]:<32} -- {task.status}"
            )
        log_statement: str = "\n".join(log_statements)

        self.logger.info("Started mission:\n%s", log_statement)

    def start_mission(
        self,
        mission: Mission,
    ) -> None:
        """Start mission

        Raises
        ------
        HTTTPException 408 Request timeout
            If there is a timeout while communicating with the state machine
        HTTPException 409 Conflict
            If the state machine is not ready to receive a mission
        HTTPException 500 Internal Server Error
            If there is an unexpected error while sending the mission to the state machine
        """
        try:
            self.logger.info(
                "Requesting to start mission:\n"
                f"  Mission ID: {mission.id}\n"
                f"  Mission Name: {mission.name}\n"
                f"  Number of Tasks: {len(mission.tasks)}"
            )
            mission_start_response = self._send_command(
                deepcopy(mission),
                self.api_events.start_mission,
            )
            if not mission_start_response.mission_started:
                self.logger.warning(
                    f"Mission failed to start - {mission_start_response.mission_not_started_reason}"
                )
                raise HTTPException(
                    status_code=HTTPStatus.CONFLICT,
                    detail=mission_start_response.mission_not_started_reason,
                )
        except EventConflictError:
            error_message = "Previous mission request is still being processed"
            self.logger.warning(error_message)
            raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)
        except EventTimeoutError:
            error_message = (
                "State machine has entered a state which cannot start a mission"
            )
            self.logger.warning(error_message)
            raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)
        except Exception as e:
            error_message = "Unexpected error while sending mission to state machine"
            self.logger.error(f"{error_message}. Exception: {e}")
            raise HTTPException(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR, detail=error_message
            )
        self.log_mission_overview(mission)
        self.logger.info("OK - Mission start successfully initiated")

    def return_home(
        self,
    ) -> None:
        """Start return home mission

        Raises
        ------
        HTTTPException 408 Request timeout
            If there is a timeout while communicating with the state machine
        HTTPException 409 Conflict
            If the state machine is not ready to receive a return home mission
        HTTPException 500 Internal Server Error
            If there is an unexpected error while sending the return home command
        """
        try:
            self._send_command(
                EmptyMessage(),
                self.api_events.return_home,
            )
        except EventConflictError:
            error_message = "Previous return home request is still being processed"
            self.logger.warning(error_message)
            raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)
        except EventTimeoutError:
            error_message = "State machine has entered a state which cannot start a return home mission"
            self.logger.warning(error_message)
            raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)
        except Exception as e:
            error_message = "Unexpected error while sending return home command"
            self.logger.error(f"{error_message}. Exception: {e}")
            raise HTTPException(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR, detail=error_message
            )
        self.logger.info("OK - Return home mission start successfully initiated")

    def pause_mission(self) -> ControlMissionResponse:
        """Pause mission

        Raises
        ------
        HTTTPException 408 Request timeout
            If there is a timeout while communicating with the state machine
        HTTPException 409 Conflict
            If the state machine is not in a state which can pause a mission
        """
        try:
            response = self._send_command(EmptyMessage(), self.api_events.pause_mission)
            if not response.success:
                self.logger.warning(
                    f"Mission failed to pause - {response.failure_reason}"
                )
                raise HTTPException(
                    status_code=HTTPStatus.CONFLICT,
                    detail=response.failure_reason,
                )
            self.logger.info("OK - Pause mission successfully initiated")
            return response
        except EventConflictError:
            error_message = "Previous pause mission request is still being processed"
            self.logger.warning(error_message)
            raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)
        except EventTimeoutError:
            error_message = (
                "State machine has entered a state which cannot pause a mission"
            )
            self.logger.warning(error_message)
            raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)
        except Exception as e:
            error_message = "Unexpected error while pausing mission"
            self.logger.error(f"{error_message}. Exception: {e}")
            raise HTTPException(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR, detail=error_message
            )

    def resume_mission(self) -> ControlMissionResponse:
        """Resume mission

        Raises
        ------
        HTTTPException 408 Request timeout
            If there is a timeout while communicating with the state machine
        HTTPException 409 Conflict
            If the state machine is not in a state which can resume a mission
        HTTPException 500 Internal Server Error
            If there is an unexpected error while resuming the mission
        """
        try:
            response = self._send_command(
                EmptyMessage(), self.api_events.resume_mission
            )
            self.logger.info("OK - Resume mission successfully initiated")
            return response
        except EventConflictError:
            error_message = "Previous resume mission request is still being processed"
            self.logger.warning(error_message)
            raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)
        except EventTimeoutError:
            error_message = "Internal Server Error - Failed to resume mission"
            self.logger.error(error_message)
            raise HTTPException(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR, detail=error_message
            )
        except Exception as e:
            error_message = "Unexpected error while resuming mission"
            self.logger.error(f"{error_message}. Exception: {e}")
            raise HTTPException(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR, detail=error_message
            )

    def stop_mission(self, mission_id: str = "") -> ControlMissionResponse:
        """Stop mission

        Raises
        ------
        HTTPException 404 Not Found
            If the mission_id was not known to Isar
        HTTPException 503 Service Unavailable
            The request was understood, but attempting to stop the mission failed
        HTTPException 408 Request timeout
            If there is a timeout while communicating with the state machine
        HTTPException 409 Conflict
            If the state machine is not in a state which can stop a mission
        HTTPException 500 Internal Server Error
            If there is an unexpected error while stopping the mission
        """
        try:
            stop_mission_response: ControlMissionResponse = self._send_command(
                mission_id, self.api_events.stop_mission
            )

            if not stop_mission_response.success:
                error_message = (
                    f"Failed to stop mission: {stop_mission_response.failure_reason}"
                )
                self.logger.error(error_message)
                raise HTTPException(
                    status_code=HTTPStatus.SERVICE_UNAVAILABLE, detail=error_message
                )
        except EventConflictError:
            error_message = "Previous stop mission request is still being processed"
            self.logger.warning(error_message)
            raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)
        except EventTimeoutError:
            error_message = (
                "State machine has entered a state which cannot stop a mission"
            )
            self.logger.warning(error_message)
            raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)
        except HTTPException as e:
            raise e
        except Exception as e:
            error_message = "Unexpected error while stopping mission"
            self.logger.error(f"{error_message}. Exception: {e}")
            raise HTTPException(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR, detail=error_message
            )
        self.logger.info("OK - Stop mission successfully initiated")
        return stop_mission_response

    def release_intervention_needed(self) -> None:
        """Release intervention needed state

        Raises
        ------
        HTTPException 409 Conflict
            If the state machine is not in intervention needed state
        HTTPException 408 Request timeout
            If there is a timeout while communicating with the state machine
        HTTPException 500 Internal Server Error
            If the intervention needed state could not be released
        """
        try:
            self._send_command(
                EmptyMessage(), self.api_events.release_intervention_needed
            )
            self.logger.info("OK - Intervention needed state released")
        except EventConflictError:
            error_message = (
                "Previous release intervention needed request is still being processed"
            )
            self.logger.warning(error_message)
            raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)
        except EventTimeoutError:
            error_message = "Cannot release intervention needed as it is not in intervention needed state"
            self.logger.warning(error_message)
            raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)
        except Exception as e:
            error_message = "Unexpected error while releasing intervention needed state"
            self.logger.error(f"{error_message}. Exception: {e}")
            raise HTTPException(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR, detail=error_message
            )

    def lock_down_robot(self) -> None:
        """Lock down robot

        Raises
        ------
        HTTPException 409 Conflict
            If the state machine is not in a state which can be locked down
        HTTPException 500 Internal Server Error
            If the robot could not be locked down
        """
        try:
            self._send_command(EmptyMessage(), self.api_events.send_to_lockdown)
            self.logger.info("OK - Robot sent into lockdown")
        except EventConflictError:
            error_message = "Previous lockdown request is still being processed"
            self.logger.warning(error_message)
            raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)
        except EventTimeoutError:
            error_message = "Cannot send robot to lockdown as it is already in lockdown"
            self.logger.warning(error_message)
            raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)
        except Exception as e:
            error_message = "Unexpected error while locking down robot"
            self.logger.error(f"{error_message}. Exception: {e}")
            raise HTTPException(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR, detail=error_message
            )

    def release_robot_lockdown(self) -> None:
        """Release robot from lockdown

        Raises
        ------
        HTTPException 409 Conflict
            If the state machine is not in lockdown
        HTTPException 500 Internal Server Error
            If the robot could not be released from lockdown
        """
        try:
            self._send_command(EmptyMessage(), self.api_events.release_from_lockdown)
            self.logger.info("OK - Robot released form lockdown")
        except EventConflictError:
            error_message = (
                "Previous release robot from lockdown request is still being processed"
            )
            self.logger.warning(error_message)
            raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)
        except EventTimeoutError:
            error_message = (
                "Cannot release robot from lockdown as it is not in lockdown"
            )
            self.logger.warning(error_message)
            raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)
        except Exception as e:
            error_message = "Unexpected error while releasing robot from lockdown"
            self.logger.error(f"{error_message}. Exception: {e}")
            raise HTTPException(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR, detail=error_message
            )

    def set_maintenance_mode(self) -> None:
        """Set maintenance mode"""
        try:
            if settings.PERSISTENT_STORAGE_CONNECTION_STRING != "":
                change_persistent_robot_state_is_maintenance_mode(
                    settings.PERSISTENT_STORAGE_CONNECTION_STRING,
                    settings.ISAR_ID,
                    value=True,
                )
            response: MaintenanceResponse = self._send_command(
                EmptyMessage(), self.api_events.set_maintenance_mode
            )
            if response.failure_reason is not None:
                self.logger.warning(response.failure_reason)
                raise HTTPException(
                    status_code=HTTPStatus.CONFLICT,
                    detail="Conflict attempting to set maintenance mode",
                )
            self.logger.info("OK - Robot sent into maintenance mode")
        except EventConflictError:
            error_message = "Previous maintenance request is still being processed"
            self.logger.warning(error_message)
            raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)
        except EventTimeoutError:
            error_message = (
                "Cannot send robot to maintenance as it is already in maintenance"
            )
            self.logger.warning(error_message)
            raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)
        except Exception as e:
            error_message = "Unexpected error while setting maintenance mode"
            self.logger.error(f"{error_message} Exception: {e}")
            raise HTTPException(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR, detail=error_message
            )

    def release_maintenance_mode(self) -> None:
        """Release robot from maintenance mode"""
        try:
            self._send_command(
                EmptyMessage(), self.api_events.release_from_maintenance_mode
            )
            if settings.PERSISTENT_STORAGE_CONNECTION_STRING != "":
                change_persistent_robot_state_is_maintenance_mode(
                    settings.PERSISTENT_STORAGE_CONNECTION_STRING,
                    settings.ISAR_ID,
                    value=False,
                )
            self.logger.info("OK - Robot released form maintenance mode")
        except EventConflictError:
            error_message = "Previous release robot from maintenance request is still being processed"
            self.logger.warning(error_message)
            raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)
        except EventTimeoutError:
            error_message = (
                "Cannot release robot from maintenance as it is not in maintenance"
            )
            self.logger.warning(error_message)
            raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)
        except Exception as e:
            error_message = "Unexpected error while releasing maintenance mode"
            self.logger.error(f"{error_message} Exception: {e}")
            raise HTTPException(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR, detail=error_message
            )

    def _send_command(self, input: T1, api_event: APIEvent[T1, T2]) -> T2:
        if not api_event.lock.acquire(blocking=False):
            raise EventConflictError("API event has already been sent")

        try:
            if api_event.request.has_event():
                self.logger.error(
                    "API request already had pending request before sending request"
                )

            if api_event.response.has_event():
                self.logger.error(
                    "API request already had response before sending request"
                )

            api_event.request.clear_event()
            api_event.response.clear_event()

            api_event.request.trigger_event(input, timeout=1)
            return api_event.response.consume_event(timeout=settings.QUEUE_TIMEOUT)
        except EventTimeoutError as e:
            self.logger.error("Queue timed out")
            api_event.request.clear_event()
            self.logger.error("No output received for command to state machine")
            raise e
        finally:
            api_event.request.clear_event()
            api_event.response.clear_event()
            api_event.lock.release()



import time
from collections import deque
from typing import List

import pytest
from fastapi import HTTPException
from pytest_mock import MockerFixture

from isar.config.settings import settings
from isar.modules import ApplicationContainer
from isar.robot.robot_status import RobotStatusThread
from isar.services.utilities.scheduling_utilities import SchedulingUtilities
from isar.state_machine.states_enum import States
from isar.storage.storage_interface import StorageInterface
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.status import MissionStatus, RobotStatus, TaskStatus
from robot_interface.models.mission.task import TakeImage, Task
from tests.test_mocks.pose import DummyPose
from tests.test_mocks.robot_interface import (
    StubRobot,
    StubRobotInitiateMissionRaisesException,
    StubRobotRobotStatusBusyIfNotHomeOrUnknownStatus,
)
from tests.test_mocks.state_machine_mocks import (
    RobotServiceThreadMock,
    StateMachineThreadMock,
    UploaderThreadMock,
)
from tests.test_mocks.task import StubTask


def test_state_machine_transitions_when_running_full_mission(
    container: ApplicationContainer,
    state_machine_thread: StateMachineThreadMock,
    robot_service_thread: RobotServiceThreadMock,
    mocker: MockerFixture,
) -> None:
    mocker.patch.object(settings, "RETURN_HOME_DELAY", 0.01)

    state_machine_thread.start()
    robot_service_thread.start()
    time.sleep(1)
    # Setting the poll interval to a lower value to ensure that the robot status is
    # updated during the mission. This value needs to be set after the robot service
    # thread has been started.
    robot_service_thread.robot_service.robot = (
        StubRobotRobotStatusBusyIfNotHomeOrUnknownStatus(
            current_state=robot_service_thread.robot_service.shared_state.state,
            initiate_mission_delay=1,
        )
    )
    robot_service_thread.robot_service.robot_status_thread.robot_status_poll_interval = (
        0.5
    )

    task_1: Task = TakeImage(
        target=DummyPose.default_pose().position, robot_pose=DummyPose.default_pose()
    )
    task_2: Task = TakeImage(
        target=DummyPose.default_pose().position, robot_pose=DummyPose.default_pose()
    )
    mission: Mission = Mission(name="Dummy mission", tasks=[task_1, task_2])

    scheduling_utilities: SchedulingUtilities = container.scheduling_utilities()
    scheduling_utilities.start_mission(mission=mission)
    time.sleep(3)  # Allow enough time to run mission and return home

    assert state_machine_thread.state_machine.transitions_list == deque(
        [
            States.UnknownStatus,
            States.AwaitNextMission,
            States.Monitor,
            States.AwaitNextMission,
            States.ReturningHome,
            States.Home,
        ]
    )


def test_state_machine_failed_dependency(
    container: ApplicationContainer,
    state_machine_thread: StateMachineThreadMock,
    robot_service_thread: RobotServiceThreadMock,
    mocker: MockerFixture,
) -> None:
    mocker.patch.object(settings, "RETURN_HOME_DELAY", 0.01)
    mocker.patch.object(settings, "RETURN_HOME_RETRY_LIMIT", 3)

    task_1: Task = TakeImage(
        target=DummyPose.default_pose().position, robot_pose=DummyPose.default_pose()
    )
    task_2: Task = TakeImage(
        target=DummyPose.default_pose().position, robot_pose=DummyPose.default_pose()
    )
    mission: Mission = Mission(name="Dummy misson", tasks=[task_1, task_2])

    mocker.patch.object(StubRobot, "task_status", return_value=TaskStatus.Failed)
    mocker.patch.object(StubRobot, "mission_status", return_value=MissionStatus.Failed)

    state_machine_thread.start()
    robot_service_thread.start()
    time.sleep(1)
    scheduling_utilities: SchedulingUtilities = container.scheduling_utilities()
    scheduling_utilities.start_mission(mission=mission)
    time.sleep(3)  # Allow the state machine to transition through the mission

    assert state_machine_thread.state_machine.transitions_list == deque(
        [
            States.UnknownStatus,
            States.AwaitNextMission,
            States.Monitor,
            States.AwaitNextMission,
            States.ReturningHome,
            States.InterventionNeeded,
        ]
    )


def test_state_machine_with_successful_collection(
    container: ApplicationContainer,
    state_machine_thread: StateMachineThreadMock,
    robot_service_thread: RobotServiceThreadMock,
    uploader_thread: UploaderThreadMock,
    mocker: MockerFixture,
) -> None:
    mocker.patch.object(StubRobot, "robot_status", return_value=RobotStatus.Home)

    storage_mock: StorageInterface = container.storage_handlers(List[StorageInterface])[
        0
    ]

    mocker.patch.object(
        RobotStatusThread, "_is_ready_to_poll_for_status", return_value=True
    )

    mission: Mission = Mission(name="Dummy misson", tasks=[StubTask.take_image()])
    scheduling_utilities: SchedulingUtilities = container.scheduling_utilities()

    mocker.patch.object(settings, "RETURN_HOME_DELAY", 0.01)
    state_machine_thread.start()
    uploader_thread.start()

    robot_service_thread.start()
    time.sleep(1)
    scheduling_utilities.start_mission(mission=mission)
    time.sleep(3)  # Allow enough time to run mission and return home

    expected_stored_items = 1
    assert len(storage_mock.stored_inspections) == expected_stored_items  # type: ignore
    assert state_machine_thread.state_machine.transitions_list == deque(
        [
            States.UnknownStatus,
            States.Home,
            States.Monitor,
            States.AwaitNextMission,
            States.ReturningHome,
            States.Home,
        ]
    )


def test_state_machine_with_unsuccessful_collection(
    container: ApplicationContainer,
    mocker: MockerFixture,
    state_machine_thread: StateMachineThreadMock,
    robot_service_thread: RobotServiceThreadMock,
    uploader_thread: UploaderThreadMock,
) -> None:
    mocker.patch.object(StubRobot, "robot_status", return_value=RobotStatus.Home)

    storage_mock: StorageInterface = container.storage_handlers(List[StorageInterface])[
        0
    ]

    mocker.patch.object(StubRobot, "get_inspection", return_value=None)

    mocker.patch.object(
        RobotStatusThread, "_is_ready_to_poll_for_status", return_value=True
    )

    mocker.patch.object(settings, "RETURN_HOME_DELAY", 0.01)
    state_machine_thread.start()
    robot_service_thread.start()
    uploader_thread.start()
    time.sleep(1)
    mission: Mission = Mission(name="Dummy misson", tasks=[StubTask.take_image()])
    scheduling_utilities: SchedulingUtilities = container.scheduling_utilities()
    scheduling_utilities.start_mission(mission=mission)
    time.sleep(3)  # Allow enough time to run mission and return home

    expected_stored_items = 0
    assert len(storage_mock.stored_inspections) == expected_stored_items  # type: ignore

    assert state_machine_thread.state_machine.transitions_list == deque(
        [
            States.UnknownStatus,
            States.Home,
            States.Monitor,
            States.AwaitNextMission,
            States.ReturningHome,
            States.Home,
        ]
    )


def test_state_machine_with_mission_start_during_return_home_without_queueing_stop_response(
    container: ApplicationContainer,
    mocker: MockerFixture,
    state_machine_thread: StateMachineThreadMock,
    robot_service_thread: RobotServiceThreadMock,
) -> None:
    mocker.patch.object(StubRobot, "robot_status", return_value=RobotStatus.Home)
    mission: Mission = Mission(name="Dummy misson", tasks=[StubTask.take_image()])
    scheduling_utilities: SchedulingUtilities = container.scheduling_utilities()
    mocker.patch.object(
        StubRobot, "mission_status", return_value=MissionStatus.InProgress
    )

    settings.FSM_SLEEP_TIME = 0

    state_machine_thread.start()
    robot_service_thread.start()
    time.sleep(1)
    scheduling_utilities.return_home()
    time.sleep(1)
    scheduling_utilities.start_mission(mission=mission)
    time.sleep(1)
    assert state_machine_thread.state_machine.transitions_list == deque(
        [
            States.UnknownStatus,
            States.Home,
            States.ReturningHome,
            States.StoppingReturnHome,
            States.Monitor,
        ]
    )
    assert (
        not state_machine_thread.state_machine.events.api_requests.start_mission.request.has_event()
    )


def test_state_machine_failed_to_initiate_mission_and_return_home(
    container: ApplicationContainer,
    state_machine_thread: StateMachineThreadMock,
    robot_service_thread: RobotServiceThreadMock,
    mocker: MockerFixture,
) -> None:
    mocker.patch.object(settings, "RETURN_HOME_DELAY", 0.01)

    robot_service_thread.robot_service.robot = StubRobotInitiateMissionRaisesException()

    task_1: Task = TakeImage(
        target=DummyPose.default_pose().position, robot_pose=DummyPose.default_pose()
    )
    task_2: Task = TakeImage(
        target=DummyPose.default_pose().position, robot_pose=DummyPose.default_pose()
    )
    mission: Mission = Mission(name="Dummy misson", tasks=[task_1, task_2])

    state_machine_thread.start()
    robot_service_thread.start()

    # TODO: check mqtt
    time.sleep(1)
    scheduling_utilities: SchedulingUtilities = container.scheduling_utilities()
    scheduling_utilities.start_mission(mission=mission)
    time.sleep(3)  # Allow the state machine to transition through the mission

    assert state_machine_thread.state_machine.transitions_list == deque(
        [
            States.UnknownStatus,
            States.AwaitNextMission,
            States.Monitor,
            States.AwaitNextMission,
            States.ReturningHome,
            States.InterventionNeeded,
        ]
    )


def test_state_machine_battery_too_low_to_start_mission(
    container: ApplicationContainer,
    state_machine_thread: StateMachineThreadMock,
    robot_service_thread: RobotServiceThreadMock,
    mocker: MockerFixture,
) -> None:
    mocker.patch.object(settings, "RETURN_HOME_DELAY", 0.01)
    state_machine_thread.start()
    mocker.patch.object(StubRobot, "robot_status", return_value=RobotStatus.Home)
    mocker.patch.object(StubRobot, "get_battery_level", return_value=10.0)
    robot_service_thread.start()
    time.sleep(1)
    task_1: Task = TakeImage(
        target=DummyPose.default_pose().position, robot_pose=DummyPose.default_pose()
    )
    task_2: Task = TakeImage(
        target=DummyPose.default_pose().position, robot_pose=DummyPose.default_pose()
    )
    mission: Mission = Mission(name="Dummy misson", tasks=[task_1, task_2])

    scheduling_utilities: SchedulingUtilities = container.scheduling_utilities()

    with pytest.raises(HTTPException) as exception_details:
        scheduling_utilities.start_mission(mission=mission)
        assert exception_details.value.status_code == 408

    assert state_machine_thread.state_machine.transitions_list == deque(
        [
            States.UnknownStatus,
            States.Home,
            States.Recharging,
        ]
    )



import json
import logging
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from queue import Empty, Queue
from threading import Event
from typing import List

from isar.config.settings import settings
from isar.models.events import Events
from isar.storage.storage_interface import (
    BlobStoragePath,
    LocalStoragePath,
    StorageException,
    StorageInterface,
    StoragePaths,
)
from robot_interface.models.inspection.inspection import (
    Inspection,
    InspectionBlob,
    InspectionValue,
)
from robot_interface.models.mission.mission import Mission
from robot_interface.telemetry.mqtt_client import MqttClientInterface
from robot_interface.telemetry.payloads import (
    InspectionResultPayload,
    InspectionValuePayload,
)
from robot_interface.utilities.json_service import EnhancedJSONEncoder


def has_empty_blob_storage_path(storage_paths: StoragePaths) -> bool:
    for path in (storage_paths.data_path, storage_paths.metadata_path):
        for value in (path.storage_account, path.blob_container, path.blob_name):
            if not (value and value.strip()):
                return True
    return False


@dataclass
class UploaderQueueItem:
    inspection: Inspection
    mission: Mission


@dataclass
class ValueItem(UploaderQueueItem):
    inspection: InspectionValue


@dataclass
class BlobItem(UploaderQueueItem):
    inspection: InspectionBlob
    storage_handler: StorageInterface
    _retry_count: int
    _next_retry_time: datetime = datetime.now(timezone.utc)

    def increment_retry(self, max_wait_time: int) -> None:
        self._retry_count += 1
        seconds_until_retry: int = min(2**self._retry_count, max_wait_time)
        self._next_retry_time = datetime.now(timezone.utc) + timedelta(
            seconds=seconds_until_retry
        )

    def get_retry_count(self) -> int:
        return self._retry_count

    def is_ready_for_upload(self) -> bool:
        return datetime.now(timezone.utc) >= self._next_retry_time

    def seconds_until_retry(self) -> int:
        return max(
            0, int((self._next_retry_time - datetime.now(timezone.utc)).total_seconds())
        )


class Uploader:
    def __init__(
        self,
        events: Events,
        storage_handlers: List[StorageInterface],
        mqtt_publisher: MqttClientInterface,
        max_wait_time: int = settings.UPLOAD_FAILURE_MAX_WAIT,
        max_retry_attempts: int = settings.UPLOAD_FAILURE_ATTEMPTS_LIMIT,
    ) -> None:
        """Initializes the uploader.

        Parameters
        ----------
        events : Events
            Events used for cross-thread communication.
        storage_handlers : List[StorageInterface]
            List of handlers for different upload options
        max_wait_time : float
            The maximum wait time between two retries (exponential backoff)
        max_retry_attempts : int
            Maximum attempts to retry an upload when it fails
        """
        self.upload_queue: Queue = events.upload_queue
        self.storage_handlers: List[StorageInterface] = storage_handlers
        self.mqtt_publisher = mqtt_publisher

        self.max_wait_time = max_wait_time
        self.max_retry_attempts = max_retry_attempts
        self._internal_upload_queue: List[UploaderQueueItem] = []

        self.signal_thread_quitting: Event = Event()

        self.logger = logging.getLogger("uploader")

    def stop(self) -> None:
        self.signal_thread_quitting.set()

    def run(self) -> None:
        self.logger.info("Started uploader")
        while not self.signal_thread_quitting.wait(0):
            inspection: Inspection
            mission: Mission
            try:
                if self._internal_upload_queue:
                    self._process_upload_queue()

                inspection, mission = self.upload_queue.get(timeout=1)

                if not mission:
                    self.logger.warning(
                        "Failed to upload missing mission from upload queue"
                    )
                    continue

                new_item: UploaderQueueItem
                if isinstance(inspection, InspectionValue):
                    new_item = ValueItem(inspection, mission)
                    self._internal_upload_queue.append(new_item)

                elif isinstance(inspection, InspectionBlob):
                    # If new item from thread queue, add one per handler to internal queue:
                    for storage_handler in self.storage_handlers:
                        new_item = BlobItem(
                            inspection, mission, storage_handler, _retry_count=-1
                        )
                        self._internal_upload_queue.append(new_item)
                else:
                    self.logger.warning(
                        f"Unable to add UploaderQueueItem as its type {type(inspection).__name__} is unsupported"
                    )
            except Empty:
                continue
            except Exception as e:
                self.logger.error(f"Unexpected error in uploader thread: {e}")
                continue

    def _upload(self, item: BlobItem) -> StoragePaths:
        inspection_paths: StoragePaths
        try:
            inspection_paths = item.storage_handler.store(
                inspection=item.inspection, mission=item.mission
            )
            self.logger.info(
                f"Storage handler: {type(item.storage_handler).__name__} "
                f"uploaded inspection {str(item.inspection.id)[:8]}"
            )
            self._internal_upload_queue.remove(item)
        except StorageException as e:
            if item.get_retry_count() < self.max_retry_attempts:
                item.increment_retry(self.max_wait_time)
                self.logger.warning(
                    f"Storage handler: {type(item.storage_handler).__name__} "
                    f"failed to upload inspection: "
                    f"{str(item.inspection.id)[:8]}. "
                    f"Retrying in {item.seconds_until_retry()}s."
                )
            else:
                self.logger.error(
                    f"Storage handler: {type(item.storage_handler).__name__} "
                    f"exceeded max retries to upload inspection: "
                    f"{str(item.inspection.id)[:8]}. Aborting upload."
                )
                self._internal_upload_queue.remove(item)
            raise e
        return inspection_paths

    def _process_upload_queue(self) -> None:
        def should_upload(_item: UploaderQueueItem) -> bool:
            if isinstance(_item, ValueItem):
                return True
            if isinstance(_item, BlobItem) and _item.is_ready_for_upload():
                return True
            return False

        ready_items: List[UploaderQueueItem] = [
            item for item in self._internal_upload_queue if should_upload(item)
        ]
        for item in ready_items:
            if isinstance(item, ValueItem):
                self._publish_inspection_value(item.inspection)
                self.logger.info(
                    f"Published value for inspection {str(item.inspection.id)[:8]}"
                )
                self._internal_upload_queue.remove(item)
            elif isinstance(item, BlobItem):
                try:
                    inspection_paths = self._upload(item)
                    if isinstance(inspection_paths.data_path, LocalStoragePath):
                        self.logger.info("Skipping publishing when using local storage")
                    elif isinstance(
                        inspection_paths.data_path, BlobStoragePath
                    ) and has_empty_blob_storage_path(inspection_paths):
                        self.logger.warning(
                            "Skipping publishing: Blob storage paths are empty for inspection %s",
                            str(item.inspection.id)[:8],
                        )
                    else:
                        self._publish_inspection_result(
                            inspection=item.inspection,
                            inspection_paths=inspection_paths,
                        )
                except StorageException:
                    pass
            else:
                self.logger.warning(
                    f"Unable to process upload item as its type {type(item).__name__} is not supported"
                )

    def _publish_inspection_value(self, inspection: InspectionValue) -> None:
        if not self.mqtt_publisher:
            return

        if not isinstance(inspection, InspectionValue):
            logging.warning(
                f"Excpected type InspectionValue but got {type(inspection).__name__} instead"
            )
            return

        payload: InspectionValuePayload = InspectionValuePayload(
            isar_id=settings.ISAR_ID,
            robot_name=settings.ROBOT_NAME,
            inspection_id=inspection.id,
            installation_code=settings.PLANT_SHORT_NAME,
            tag_id=inspection.metadata.tag_id,
            inspection_type=type(inspection).__name__,
            inspection_description=inspection.metadata.inspection_description,
            value=inspection.value,
            unit=inspection.unit,
            x=inspection.metadata.robot_pose.position.x,
            y=inspection.metadata.robot_pose.position.y,
            z=inspection.metadata.robot_pose.position.z,
            timestamp=inspection.metadata.start_time,
        )
        self.mqtt_publisher.publish(
            topic=settings.TOPIC_ISAR_INSPECTION_VALUE,
            payload=json.dumps(payload, cls=EnhancedJSONEncoder),
            qos=1,
            retain=True,
        )

    def _publish_inspection_result(
        self,
        inspection: InspectionBlob,
        inspection_paths: StoragePaths[BlobStoragePath],
    ) -> None:
        """Publishes the reference of the inspection result to the MQTT Broker
        along with the analysis type
        """
        if not self.mqtt_publisher:
            return

        payload: InspectionResultPayload = InspectionResultPayload(
            isar_id=settings.ISAR_ID,
            robot_name=settings.ROBOT_NAME,
            inspection_id=inspection.id,
            blob_storage_data_path=inspection_paths.data_path,
            blob_storage_metadata_path=inspection_paths.metadata_path,
            installation_code=settings.PLANT_SHORT_NAME,
            tag_id=inspection.metadata.tag_id,
            inspection_type=type(inspection).__name__,
            inspection_description=inspection.metadata.inspection_description,
            timestamp=inspection.metadata.start_time,
        )
        self.mqtt_publisher.publish(
            topic=settings.TOPIC_ISAR_INSPECTION_RESULT,
            payload=json.dumps(payload, cls=EnhancedJSONEncoder),
            qos=1,
            retain=True,
        )



from isar.models.events import Event, Events


class TestEvents:
    def test_events(self) -> None:
        events: Events = Events()
        assert events.api_requests.start_mission is not None
        assert (
            events.api_requests.start_mission.request is not None
            and events.api_requests.start_mission.request.maxsize == 1
        )
        assert (
            events.api_requests.start_mission.response is not None
            and events.api_requests.start_mission.response.maxsize == 1
        )
        assert events.api_requests.stop_mission is not None
        assert (
            events.api_requests.stop_mission.request is not None
            and events.api_requests.stop_mission.request.maxsize == 1
        )
        assert (
            events.api_requests.stop_mission.response is not None
            and events.api_requests.stop_mission.response.maxsize == 1
        )


def test_staus_queue_empty() -> None:
    status_event: Event = Event("test")
    assert status_event.check() is None


def test_status_event_check() -> None:
    status_event: Event = Event("test")
    status_event.update("Test")
    assert status_event.check() == "Test"
    assert status_event._qsize() == 1


def test_status_event_update() -> None:
    status_event: Event = Event("test")
    status_event.update("Test")
    status_event.update("New Test")
    assert status_event._qsize() == 1
    assert status_event.check() == "New Test"



import json
import time
from datetime import datetime, timezone
from queue import Queue

from isar.config.settings import settings
from isar.services.service_connections.mqtt.mqtt_client import props_expiry
from robot_interface.telemetry.mqtt_client import MqttPublisher
from robot_interface.telemetry.payloads import RobotHeartbeatPayload
from robot_interface.utilities.json_service import EnhancedJSONEncoder


class RobotHeartbeatPublisher:
    def __init__(self, mqtt_queue: Queue):
        self.mqtt_publisher: MqttPublisher = MqttPublisher(mqtt_queue=mqtt_queue)

    def run(self) -> None:
        while True:
            payload: RobotHeartbeatPayload = RobotHeartbeatPayload(
                isar_id=settings.ISAR_ID,
                robot_name=settings.ROBOT_NAME,
                timestamp=datetime.now(timezone.utc),
            )

            self.mqtt_publisher.publish(
                topic=settings.TOPIC_ISAR_ROBOT_HEARTBEAT,
                payload=json.dumps(payload, cls=EnhancedJSONEncoder),
                retain=True,
                properties=props_expiry(settings.MQTT_ROBOT_HEARTBEAT_EXPIRY),
            )

            time.sleep(settings.ROBOT_HEARTBEAT_PUBLISH_INTERVAL)



from typing import Tuple

from isar.state_machine.states_enum import States


def stub_status(
    mission_in_progress: bool = True,
    current_state: States = States.Home,
) -> Tuple[bool, States]:
    return mission_in_progress, current_state



import logging
from http import HTTPStatus

from fastapi import Body, HTTPException
from opentelemetry import trace

from isar.apis.models.models import (
    ControlMissionResponse,
    StartMissionResponse,
    TaskResponse,
)
from isar.apis.models.start_mission_definition import (
    MissionFormatError,
    StartMissionDefinition,
    StopMissionDefinition,
    to_isar_mission,
)
from isar.config.settings import robot_settings
from isar.services.utilities.scheduling_utilities import SchedulingUtilities
from isar.state_machine.states_enum import States
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.task import TASKS, InspectionTask

tracer = trace.get_tracer(__name__)


class SchedulingController:
    def __init__(
        self,
        scheduling_utilities: SchedulingUtilities,
    ):
        self.scheduling_utilities: SchedulingUtilities = scheduling_utilities
        self.logger = logging.getLogger("api")

    @tracer.start_as_current_span("start_mission")
    def start_mission(
        self,
        mission_definition: StartMissionDefinition = Body(
            default=None,
            embed=True,
            title="Mission Definition",
            description="Description of the mission in json format",
        ),
    ) -> StartMissionResponse:
        self.logger.info("Received request to start new mission")

        if not mission_definition:
            error_message_no_mission_definition: str = (
                "Unprocessable entity - 'mission_definition' empty or invalid"
            )
            self.logger.error(error_message_no_mission_definition)
            raise HTTPException(
                status_code=HTTPStatus.UNPROCESSABLE_ENTITY,
                detail=error_message_no_mission_definition,
            )

        state: States = self.scheduling_utilities.get_state()
        self.scheduling_utilities.verify_state_machine_ready_to_receive_mission(state)

        try:
            mission: Mission = to_isar_mission(
                start_mission_definition=mission_definition
            )
        except MissionFormatError as e:
            error_message = f"Bad Request - Cannot create ISAR mission: {e}"
            self.logger.warning(error_message)
            raise HTTPException(
                status_code=HTTPStatus.BAD_REQUEST,
                detail=error_message,
            )

        self.scheduling_utilities.verify_robot_capable_of_mission(
            mission=mission, robot_capabilities=robot_settings.CAPABILITIES
        )

        self.logger.info("Starting mission: %s", mission.id)
        self.scheduling_utilities.start_mission(mission=mission)
        return self._api_response(mission)

    @tracer.start_as_current_span("return_home")
    def return_home(self) -> None:
        self.logger.info("Received request to return home")

        state: States = self.scheduling_utilities.get_state()
        self.scheduling_utilities.verify_state_machine_ready_to_receive_return_home_mission(
            state
        )

        self.scheduling_utilities.return_home()

    @tracer.start_as_current_span("pause_mission")
    def pause_mission(self) -> ControlMissionResponse:
        self.logger.info("Received request to pause current mission")

        state: States = self.scheduling_utilities.get_state()

        if state not in [
            States.Monitor,
            States.ReturningHome,
        ]:
            error_message = (
                f"Conflict - Pause command received in invalid state - State: {state}"
            )
            self.logger.warning(error_message)
            raise HTTPException(
                status_code=HTTPStatus.CONFLICT,
                detail=error_message,
            )

        pause_mission_response: ControlMissionResponse = (
            self.scheduling_utilities.pause_mission()
        )
        return pause_mission_response

    @tracer.start_as_current_span("resume_mission")
    def resume_mission(self) -> ControlMissionResponse:
        self.logger.info("Received request to resume current mission")

        state: States = self.scheduling_utilities.get_state()

        if state not in [States.Paused, States.ReturnHomePaused]:
            error_message = (
                f"Conflict - Resume command received in invalid state - State: {state}"
            )
            self.logger.warning(error_message)
            raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)

        resume_mission_response: ControlMissionResponse = (
            self.scheduling_utilities.resume_mission()
        )
        return resume_mission_response

    @tracer.start_as_current_span("stop_mission")
    def stop_mission(
        self,
        mission_id: StopMissionDefinition = Body(
            default=None,
            embed=True,
            title="Mission ID to stop",
            description="The mission ID of the mission being stopped, in json format",
        ),
    ) -> ControlMissionResponse:

        self.logger.info("Received request to stop current mission")

        state: States = self.scheduling_utilities.get_state()

        if (
            state == States.UnknownStatus
            or state == States.Stopping
            or state == States.BlockedProtectiveStop
            or state == States.Offline
            or state == States.Home
            or state == States.ReturningHome
        ):
            error_message = (
                f"Conflict - Stop command received in invalid state - State: {state}"
            )
            self.logger.warning(error_message)
            raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)

        stop_mission_response: ControlMissionResponse = (
            self.scheduling_utilities.stop_mission(mission_id.mission_id)
        )
        return stop_mission_response

    @tracer.start_as_current_span("release_intervention_needed")
    def release_intervention_needed(self) -> None:
        self.logger.info("Received request to release intervention needed state")

        state: States = self.scheduling_utilities.get_state()

        if state != States.InterventionNeeded:
            error_message = f"Conflict - Release intervention needed command received in invalid state - State: {state}"
            self.logger.warning(error_message)
            raise HTTPException(
                status_code=HTTPStatus.CONFLICT,
                detail=error_message,
            )

        self.scheduling_utilities.release_intervention_needed()
        self.logger.info("Released intervention needed state successfully")

    @tracer.start_as_current_span("lockdown")
    def lockdown(self) -> None:
        self.logger.info("Received request to lockdown robot")

        state: States = self.scheduling_utilities.get_state()

        if state == States.Lockdown:
            error_message = "Conflict - Lockdown command received in lockdown state"
            self.logger.warning(error_message)
            raise HTTPException(
                status_code=HTTPStatus.CONFLICT,
                detail=error_message,
            )

        self.scheduling_utilities.lock_down_robot()
        self.logger.info("Lockdown started successfully")

    @tracer.start_as_current_span("release_lockdown")
    def release_lockdown(self) -> None:
        self.logger.info("Received request to release robot lockdown")

        state: States = self.scheduling_utilities.get_state()

        if state != States.Lockdown:
            error_message = f"Conflict - Release lockdown command received in invalid state - State: {state}"
            self.logger.warning(error_message)
            raise HTTPException(
                status_code=HTTPStatus.CONFLICT,
                detail=error_message,
            )

        self.scheduling_utilities.release_robot_lockdown()
        self.logger.info("Released lockdown successfully")

    @tracer.start_as_current_span("maintenance_mode")
    def set_maintenance_mode(self) -> None:
        self.logger.info("Received request to set maintenance_mode")

        state: States = self.scheduling_utilities.get_state()

        if state == States.Maintenance or state == States.StoppingDueToMaintenance:
            message = f"Conflict - Call to set maintenance mode was given while in state {state}."
            self.logger.info(message)
            raise HTTPException(
                status_code=HTTPStatus.CONFLICT,
                detail=message,
            )

        self.scheduling_utilities.set_maintenance_mode()
        self.logger.info("Maintenance mode has been set")

    @tracer.start_as_current_span("release_maintenance_mode")
    def release_maintenance_mode(self) -> None:
        self.logger.info("Received request to release robot from maintenance mode")

        state: States = self.scheduling_utilities.get_state()

        if state != States.Maintenance:
            message = f"Conflict - Release maintenance mode command received in invalid state - State: {state}"
            self.logger.info(message)
            raise HTTPException(
                status_code=HTTPStatus.CONFLICT,
                detail=message,
            )

        self.scheduling_utilities.release_maintenance_mode()
        self.logger.info("Maintenance mode successfully released")

    def _api_response(self, mission: Mission) -> StartMissionResponse:
        return StartMissionResponse(
            id=mission.id,
            tasks=[self._task_api_response(task) for task in mission.tasks],
        )

    def _task_api_response(self, task: TASKS) -> TaskResponse:
        if isinstance(task, InspectionTask):
            inspection_id = task.inspection_id
        else:
            inspection_id = None

        return TaskResponse(
            id=task.id, tag_id=task.tag_id, inspection_id=inspection_id, type=task.type
        )



import json
from dataclasses import asdict, is_dataclass
from datetime import date, datetime
from enum import Enum
from typing import Any, Iterator
from uuid import UUID

import numpy as np
from alitra import Orientation
from pydantic import BaseModel


class EnhancedJSONEncoder(json.JSONEncoder):
    """
    Custom JSONEncoder with the ability to encode dataclasses.
    """

    def default(self, o: Any) -> Any:
        if isinstance(o, BaseModel):
            dump = getattr(o, "model_dump", None)
            if callable(dump):
                return dump()
            return o.__dict__
        if is_dataclass(o):
            return asdict(o)  # type: ignore
        if isinstance(o, UUID):
            return str(o)
        if isinstance(o, Orientation):
            return o.__dict__
        if isinstance(o, np.ndarray):
            return o.tolist()
        if isinstance(o, datetime):
            return o.isoformat()
        if isinstance(o, date):
            return o.isoformat()
        if isinstance(o, Enum):
            return o.value
        if isinstance(o, bytes):
            return "<<non-serializable: bytes>>"
        if isinstance(o, Iterator):
            return "<<non-serializable: iterator>>"
        return super().default(o)
